
from __future__ import print_function

import importlib

import muteria.common.mix as common_mix

from muteria.drivers.criteria import TestCriteria

import muteria.drivers.optimizers.criteriatestexecution.tools as crit_opt

class CriteriaOptimizers(common_mix.EnumAutoName):
    # Default (No Optimization)
    NO_OPTIMIZATION = importlib.import_module(".default", \
                                                package=crit_opt.__name__
                                            ).CriteriaTestExecutionOptimizer

    # Strong Mutation
    SM_OPTIMIZED_BY_WM = importlib.import_module(\
                                        ".strongmutation_by_weakmutation", \
                                        package=crit_opt.__name__
                                    ).CriteriaTestExecutionOptimizer
    SM_OPTIMIZED_BY_MCOV = importlib.import_module(\
                                        ".strongmutation_by_mutantcoverage", \
                                        package=crit_opt.__name__
                                    ).CriteriaTestExecutionOptimizer

    def get_optimizer(self):
        return self.get_field_value()
    #~ def get_optimizer():
#~ class CriteriaOptimizers

# TODO: Make a function to check that the optimizer specified is for the right criterion
def check_is_right_optimizer(criterion, optimizer):
    """ Check that the optimizer is fit for the criterion
        :return: True if fit, False otherwise
    """
    dat = {
        TestCriteria.STRONG_MUTATION: {
            CriteriaOptimizers.SM_OPTIMIZED_BY_MCOV,
            CriteriaOptimizers.SM_OPTIMIZED_BY_WM,
        }
    }

    # Check
    if criterion in dat:
        if optimizer in dat[criterion]:
            return True
    return False
#~ def check_right_optimizer()


#
# [LICENCE]
#
""" Base test criteria test execution optimizer class (For a single criterion)

    Each file that implement a optimizer must have the class 
    `CriteriaTestExecutionOptimizer` that implements the base class bellow
"""

from __future__ import print_function
import os
import sys
import logging
import abc

import muteria.common.mix as common_mix

ERROR_HANDLER = common_mix.ErrorHandler

class BaseCriteriaTestExecutionOptimizer(abc.ABC):
    '''
    '''

    def __init__(self, config, explorer, criterion, **kwargs):
        # Set Direct Arguments Variables
        self.config = config
        self.explorer = explorer
        self.criterion = criterion

        self.test_objective_ordered_list = None
        self.pointer = None
        self.test_objective_to_test_execution_optimizer = None
    #~ def __init__()

    def get_test_execution_optimizer(self, test_objective):
        """ Get an initialized test execution optimizer 
            (the user should not reset)
        """
        ERROR_HANDLER.assert_true(test_objective in \
                            self.test_objective_to_test_execution_optimizer, \
                                                    "Invalid test objective")
        return self.test_objective_to_test_execution_optimizer[test_objective]
    #~ def get_test_execution_optimizer()

    def has_next_test_objective (self):
        return self.pointer < len(self.test_objective_ordered_list)
    #~ def has_next_test_objective()

    def get_next_test_objective (self):
        ERROR_HANDLER.assert_true(\
                                self.test_objective_ordered_list is not None, \
                                'uninitialized test_objective_ordered_list')
        if self.has_next_test_objective():
            ret = self.test_objective_ordered_list[self.pointer]
            self.pointer += 1
        else:
            ret = None
        return ret
    #~ def get_next_test_objective()

    def select_test_objectives(self, test_objective_list, proportion_number, \
                                                        is_proportion=True):
        """ Apply Mutant Selection
        """
        ERROR_HANDLER.assert_true(\
                self.test_objective_ordered_list is not None, \
                                    'uninitialized test_objective_order_list')
        ERROR_HANDLER.assert_true(proportion_number > 0, \
                                        'proportion_number must be positive')
        if is_proportion:
            ERROR_HANDLER.assert_true(proportion_number <= 100, \
                                        'proportion must be <= 100')
            proportion_number = round(len(self.test_objective_ordered_list) * \
                                                    proportion_number / 100.0)
            proportion_number = max(1, int(proportion_number))
        return self.test_objective_ordered_list[:proportion_number]
    #~ def select_test_objectives()

    def feedback (self, test_objective, test_to_verdict, **kwargs):
        """ Possibly get feedback from past executions
            Override this if needed
        """
    #~ def feedback()

    #######################################################################
    ##################### Methods to implement ############################
    #######################################################################

    @classmethod
    @abc.abstractclassmethod
    def installed(cls, custom_binary_dir=None):
        """ Check that the tool is installed
            :return: bool reprenting whether the tool is installed or not 
                    (executable accessible on the path)
                    - True: the tool is installed and works
                    - False: the tool is not installed or do not work
        """
        print ("!!! Must be implemented in child class !!!")
    #~ def installed()

    @abc.abstractmethod
    def reset (self, toolalias, test_objective_list, test_list, **kwargs):
        """ Reset the optimizer 
            (compute test objective ordering and their test orderings)
        """
        print ("!!! Must be implemented in child class !!!")
    #~ def reset()
#~ class BaseCriteriaTestExecutionOptimizer



from __future__ import print_function

import os
import sys
import re
import shutil
import shlex
import logging
import subprocess

import muteria.common.mix as common_mix
import muteria.common.fs as common_fs

from muteria.repositoryandcode.codes_convert_support import CodeFormats
from muteria.repositoryandcode.callback_object import DefaultCallbackObject

from muteria.drivers.criteria.base_testcriteriatool import BaseCriteriaTool
from muteria.drivers.criteria import TestCriteria
from muteria.drivers import DriversUtils
from muteria.drivers.criteria.criteria_info import MutantsInfoObject

ERROR_HANDLER = common_mix.ErrorHandler

class CriteriaToolMart(BaseCriteriaTool):
    def __init__(self, *args, **kwargs):
        BaseCriteriaTool.__init__(self, *args, **kwargs)
        self.instrumentation_details = os.path.join(\
                    self.instrumented_code_storage_dir, '.instru.meta.json')
        self.wm_res_log_file = \
                    "label_log."+TestCriteria.WEAK_MUTATION.get_field_value()
        self.mcov_res_log_file = \
                    "label_log."+TestCriteria.MUTANT_COVERAGE.get_field_value()
        self.mutant_data = os.path.join(self.instrumented_code_storage_dir,\
                                                            "mutant_data")
        self.mart_out = os.path.join(self.mutant_data, 'mart-out-0')
        self.separate_muts_folder_name = 'mutants.out'
        self.separate_muts_dir = os.path.join(self.mart_out, \
                                                self.separate_muts_folder_name)
        self.archive_separated = True
    #~ def __init__()

    def _get_default_params(self):
        bool_params = {
            '-keep-mutants-bc': None,
            '-no-COV': None,
            '-no-Meta': None,
            '-no-WM': None, 
            '-no-mutant-info': None,
            '-print-preTCE-Meta': None,
            '-write-mutants': None,
        }
        key_val_params = {
            '-linking-flags': None,
            '-mutant-config': None,
            '-mutant-scope': None,
        }
        return bool_params, key_val_params
    #~ def _get_default_params()

    @classmethod
    def installed(cls, custom_binary_dir=None):
        """ Check that the tool is installed
            :return: bool reprenting whether the tool is installed or not 
                    (executable accessible on the path)
                    - True: the tool is installed and works
                    - False: the tool is not installed or do not work
        """
        for prog in ('mart',):
            if custom_binary_dir is not None:
                prog = os.path.join(custom_binary_dir, prog)
            if not DriversUtils.check_tool(prog=prog, args_list=['--version'],\
                                                    expected_exit_codes=[0,1]):
                return False
        return True
    #~ def installed()

    @classmethod
    def _get_meta_instrumentation_criteria(cls):
        """ Criteria where all elements are instrumented in same file
            :return: list of citeria
        """
        return [
                TestCriteria.MUTANT_COVERAGE,
                TestCriteria.WEAK_MUTATION,
               ]
    #~ def _get_meta_instrumentation_criteria()

    @classmethod
    def _get_separated_instrumentation_criteria(cls):
        """ Criteria where all elements are instrumented in different files
            :return: list of citeria
        """
        return [TestCriteria.STRONG_MUTATION]
    #~ def _get_separated_instrumentation_criteria()

    def get_instrumented_executable_paths_map(self, enabled_criteria):
        crit_to_exes_map = {}
        obj = common_fs.loadJSON(self.instrumentation_details)
        #exes = [p for _, p in list(obj.items())]
        for c, c_exes in list(obj.items()):
            for k in c_exes:
                c_exes[k] = os.path.join(self.mart_out, c_exes[k])

        for criterion in enabled_criteria:
            ERROR_HANDLER.assert_true(criterion.get_str() in obj, 
                            'criterion was not enabled during instrumentation'
                            ' with Mart: {}. Eneble it and run again'.format(\
                                                criterion.get_str()), __file__)
            crit_to_exes_map[criterion] = obj[criterion.get_str()]
        return crit_to_exes_map
    #~ def get_instrumented_executable_paths_map()

    def get_criterion_info_object(self, criterion):
        try:
            return self.mutant_info_object
        except AttributeError:
            minf_obj = MutantsInfoObject()
            mart_inf = os.path.join(self.mart_out, 'mutantsInfos.json')
            fdupes = os.path.join(self.mart_out, 'fdupes_duplicates.json')
            #in_mem_tce = os.path.join(self.mart_out, 'equidup-mutantsInfos.json')

            mart_inf_obj = common_fs.loadJSON(mart_inf)
            #in_mem_tce_obj = common_fs.loadJSON(in_mem_tce)
            if os.path.isfile(fdupes):
                fduped_obj = common_fs.loadJSON(fdupes)

                # remove equivalent et duplicates
                for mid, dups in list(fduped_obj.items()):
                    for d_mid in dups:
                        del mart_inf_obj[d_mid]

            # Add elements
            for mid, info in list(mart_inf_obj.items()):
                minf_obj.add_element(mid, mutant_type=info['Type'], \
                                        mutant_locs=info['SrcLoc'], \
                                        mutant_function_name=info['FuncName'],\
                                        IRPosInFunc=info['IRPosInFunc'])
            self.mutant_info_object = minf_obj
            return minf_obj
    #~ def get_criterion_info_object()

    def _get_single_exe_filename(self, criterion):
        try:
            return self.single_exe_filename
        except AttributeError:
            inst_path_map = \
                        self.get_instrumented_executable_paths_map([criterion])
            tmp = inst_path_map[criterion]
            r_file = list(tmp.keys())[0]
            filename = os.path.splitext(os.path.basename(tmp[r_file]))[0]
            self.single_exe_filename = {r_file: filename}
            return self.single_exe_filename.copy()
    #~ def _get_single_exe_filename()

    def _get_criterion_element_executable_path(self, criterion, element_id):
#        self.sm_separate_exes = 
        ERROR_HANDLER.assert_true(self.get_criterion_info_object(criterion).\
                                            has_element(element_id),\
                        "Inexistant mutant id: "+element_id, __file__)
        
        rel_names = []
        mut_code = {}
        for k,v in self._get_single_exe_filename(criterion).items():
            mut_code[k] = os.path.join(self.separate_muts_dir, element_id, v) 
            rel_names.append(os.path.join(self.separate_muts_folder_name, \
                                                                element_id, v))
        # If archiving
        if self.archive_separated:
            archive_path = common_fs.TarGz.get_archive_filename_of(\
                                                        self.separate_muts_dir)
            ERROR_HANDLER.assert_true(os.path.isfile(archive_path), \
                                    "Archived separated mutant file missing",\
                                    __file__)
            if os.path.isdir(self.separate_muts_dir):
                shutil.rmtree(self.separate_muts_dir)
            # Extract the selected
            for arch_name in rel_names:
                err_msg = common_fs.TarGz.extractFromArchive(archive_path, \
                                                                    arch_name)
                ERROR_HANDLER.assert_true(err_msg is None, \
                            "failed to extract, err: "+str(err_msg), __file__)
        return mut_code
    #~ def _get_criterion_element_executable_path()

    def _get_criterion_element_environment_vars(self, criterion, element_id):
        '''
            return: python dictionary with environment variable as key
                     and their values as value (all strings)
        '''
        return None
    #~ def _get_criterion_element_environment_vars()

    def _get_criteria_environment_vars(self, result_dir_tmp, enabled_criteria):
        '''
        return: python dictionary with environment variable as key
                     and their values as value (all strings)
        '''
        res = {}
        for c in enabled_criteria:
            if c == TestCriteria.WEAK_MUTATION:
                res[c] = {'MART_WM_LOG_OUTPUT': os.path.join(result_dir_tmp, \
                                                        self.wm_res_log_file)}
            elif c == TestCriteria.MUTANT_COVERAGE:
                res[c] = {'MART_WM_LOG_OUTPUT': os.path.join(result_dir_tmp, \
                                                    self.mcov_res_log_file)}
            else:
                res[c] = None
        return res
    #~ def _get_criteria_environment_vars()

    def _collect_temporary_coverage_data(self, criteria_name_list, \
                                            test_execution_verdict, \
                                            used_environment_vars, \
                                            result_dir_tmp, \
                                            testcase):
        ''' Get the list of mutants id weakly killed or covered
        '''
        pass
    #~ def _collect_temporary_coverage_data()

    def _extract_coverage_data_of_a_test(self, enabled_criteria, \
                                    test_execution_verdict, result_dir_tmp):
        ''' read json files and extract data
            return: the dict of criteria with covering count
        '''
        ERROR_HANDLER.assert_true(TestCriteria.STRONG_MUTATION not in \
                                                        enabled_criteria, \
                            "SM metamutant run not yet supported", __file__)

        def extract_covered(filename, criterion):
            mutant_id_set = set(self.get_criterion_info_object(criterion).\
                                                        get_elements_list())
            cov_res = {
                    m: common_mix.GlobalConstants.ELEMENT_NOTCOVERED_VERDICT\
                                                    for m in mutant_id_set}
            if os.path.isfile(filename):
                with open(filename) as f:
                    for line in f:
                        mut_id = line.strip()
                        # use if because mart currently do not update 
                        # WM and MCOV after fdupes TCE
                        if mut_id in mutant_id_set: 
                            cov_res[mut_id] = common_mix.GlobalConstants\
                                                    .ELEMENT_COVERED_VERDICT
            return cov_res
        #~ def extract_covered()

        res = {}
        if TestCriteria.WEAK_MUTATION in enabled_criteria:
            res[TestCriteria.WEAK_MUTATION] = extract_covered(\
                        os.path.join(result_dir_tmp, self.wm_res_log_file), \
                                                    TestCriteria.WEAK_MUTATION)
        if TestCriteria.MUTANT_COVERAGE in enabled_criteria:
            res[TestCriteria.MUTANT_COVERAGE] = extract_covered(\
                        os.path.join(result_dir_tmp, self.mcov_res_log_file), \
                                                TestCriteria.MUTANT_COVERAGE)

        return res
    #~ def _extract_coverage_data_of_a_test()

    def _do_instrument_code (self, exe_path_map, \
                                        code_builds_factory, \
                                        enabled_criteria, parallel_count=1):
        # Setup
        if os.path.isdir(self.instrumented_code_storage_dir):
            shutil.rmtree(self.instrumented_code_storage_dir)
        os.mkdir(self.instrumented_code_storage_dir)
        if os.path.isdir(self.mutant_data):
            shutil.rmtree(self.mutant_data)
        os.mkdir(self.mutant_data)

        prog = 'mart'
        if self.custom_binary_dir is not None:
            prog = os.path.join(self.custom_binary_dir, prog)
            ERROR_HANDLER.assert_true(os.path.isfile(prog), \
                            "The tool {} is missing from the specified dir {}"\
                                        .format(os.path.basename(prog), \
                                            self.custom_binary_dir), __file__)

        # get llvm compiler path
        ret, out, err = DriversUtils.execute_and_get_retcode_out_err(\
                                                        prog, ['--version'])
        llvm_compiler_path = None
        for line in out.splitlines():
            line = line.strip()
            if line.startswith('LLVM tools dir:'):
                llvm_compiler_path = line.split()[3]#[1:-1]
                break
        
        ERROR_HANDLER.assert_true(llvm_compiler_path is not None, \
                                'Problem getting llvm path for mart', __file__)

        # Build into LLVM
        back_llvm_compiler = 'clang'
        rel_path_map = {}
        exes, _ = code_builds_factory.repository_manager.\
                                                    get_relative_exe_path_map()
        for exe in exes:
            filename = os.path.basename(exe)
            rel_path_map[exe] = os.path.join(self.mutant_data, filename)
        pre_ret, ret, post_ret = code_builds_factory.transform_src_into_dest(\
                        src_fmt=CodeFormats.C_SOURCE,\
                        dest_fmt=CodeFormats.LLVM_BITCODE,\
                        src_dest_files_paths_map=rel_path_map,\
                        compiler=back_llvm_compiler, flags_list=['-g'], \
                        clean_tmp=True, reconfigure=True, \
                        llvm_compiler_path=llvm_compiler_path)
        if ret == common_mix.GlobalConstants.COMMAND_FAILURE: #TEST_EXECUTION_ERROR:
            ERROR_HANDLER.error_exit("Program {}.".format(\
                                'LLVM (clang) built problematic'), __file__)

        # Update exe_map to reflect bitcode extension
        rel2bitcode = {}
        for r_file, b_file in list(rel_path_map.items()):
            bc = b_file+'.bc'
            ERROR_HANDLER.assert_true(os.path.isfile(bc), \
                                    "Bitcode file not existing: "+bc, __file__)
            rel2bitcode[r_file] = bc

        ERROR_HANDLER.assert_true(len(rel_path_map) == 1, \
                            "Support single bitcode module for now", __file__)

        bitcode_file = rel2bitcode[list(rel2bitcode.keys())[0]]

        # mart params
        bool_param, k_v_params = self._get_default_params()
        if TestCriteria.STRONG_MUTATION in enabled_criteria:
            bool_param['-write-mutants'] = True
        else:
            self.archive_separated = False
        if k_v_params['-mutant-scope'] is None:
            src_list_scope_file = os.path.join(self.mutant_data, \
                                                        'mutation-scope.json')
            scope_obj = {'Source-Files': [], 'Functions': []}
            _, src2obj = self.code_builds_factory.repository_manager\
                                                .get_relative_exe_path_map()
            for src, _ in src2obj.items():
                scope_obj['Source-Files'].append(src)
            if len(scope_obj['Source-Files']) > 0:
                common_fs.dumpJSON(scope_obj, src_list_scope_file, pretty=True)
                k_v_params['-mutant-scope'] = src_list_scope_file
        
        # Consider user custom
        extra_linking_flags = ""
        uc = self.config.get_tool_user_custom()
        pre_args = []
        post_args = []
        if uc is not None:
            pre_bc_cmd = uc.PRE_TARGET_CMD_ORDERED_FLAGS_LIST
            post_bc_cmd = uc.POST_TARGET_CMD_ORDERED_FLAGS_LIST
            for _args, _cmd in [(pre_args, pre_bc_cmd), \
                                                    (post_args, post_bc_cmd)]:
                if _cmd is not None:
                    for tup in _cmd:
                        key = tup[0][1:] if tup[0].startswith('--') else tup[0]
                        bool_disabled = False
                        if key in k_v_params:
                            del k_v_params[key]
                        if key in bool_param:
                            del bool_param[key]
                            if len(tup) == 2:
                                ERROR_HANDLER.assert_true(\
                                            type(tup[1]) == bool, \
                                            "bool arg is either enabled or "
                                        "disabled. val must be bool", __file__)
                                if not tup[1]:
                                    bool_disabled = True
                                tup = tup[:1]
                            else:
                                ERROR_HANDLER.assert_true(len(tup) == 1, \
                                        "Invalid bool param: "+key, __file__)
                        if not bool_disabled:
                            _args.extend(list(tup))
                        if key == '-linking-flags':
                            ERROR_HANDLER.assert_true(len(tup) == 2, \
                                    "invalid linking-flags for mart", __file__)
                            extra_linking_flags = tup[1]

        args = [bp for bp, en in list(bool_param.items()) if en]
        for k,v in list(k_v_params.items()):
            if v is not None:
                args += [k,v]
        args.extend(pre_args)
        args.append(bitcode_file)
        args.extend(post_args)

        # make sure that the bitcode file can be compiled to native code
        ret, out, err = DriversUtils.execute_and_get_retcode_out_err('clang', \
                                extra_linking_flags.split() + [bitcode_file, \
                                                '-o', bitcode_file+'.native'])
        if ret not in (0,):
            logging.error(out)
            ERROR_HANDLER.error_exit("'Mart' cannot compile bitcode"
                            + "into native code, you need to specify"
                            + "linking flags (based on the above error).", \
                                                                    __file__)

        
        # Execute Mart
        cwd = os.getcwd()
        os.chdir(self.mutant_data)
        ret, out, err = DriversUtils.execute_and_get_retcode_out_err(\
                                                                    prog, args)
        os.chdir(cwd)

        if (ret != 0):
            logging.error(out)
            logging.error(err)
            logging.error("\n>> CMD: " + " ".join([prog]+args) + '\n')
            ERROR_HANDLER.error_exit("mart failed (ran from {})".format(\
                                                self.mutant_data), __file__)
        
        # write down the rel_path_map
        ERROR_HANDLER.assert_true(not os.path.isfile(\
                self.instrumentation_details), "must not exist here", __file__)
        store_obj = {c.get_str(): {} for c in enabled_criteria}
        for k,v in list(rel_path_map.items()):
            exe_file = os.path.basename(v)
            if TestCriteria.WEAK_MUTATION in enabled_criteria:
                crit_str = TestCriteria.WEAK_MUTATION.get_str()
                store_obj[crit_str][k] = exe_file+'.WM'
            if TestCriteria.MUTANT_COVERAGE in enabled_criteria:
                crit_str = TestCriteria.MUTANT_COVERAGE.get_str()
                store_obj[crit_str][k] = exe_file+'.COV'
            if TestCriteria.STRONG_MUTATION in enabled_criteria:
                crit_str = TestCriteria.STRONG_MUTATION.get_str()
                store_obj[crit_str][k] = exe_file+'.MetaMu'
        common_fs.dumpJSON(store_obj, self.instrumentation_details)

        # Archive separated if on
        if self.archive_separated:
            err_msg = common_fs.TarGz.compressDir(self.separate_muts_dir, \
                                                    remove_in_directory=True)
            ERROR_HANDLER.assert_true(err_msg is None,\
                                "Compression failed: "+str(err_msg), __file__)
    #~ def _do_instrument_code()

    ## Extra functions for mart
    def get_test_gen_metamutant_bc(self):
        ERROR_HANDLER.assert_true(os.path.isfile(self.instrumentation_details), \
		 "Are you sure mutant generation with Mart was ran?"
                 "This is needed before meta mu bc is called", __file__)
        crit2file = self.get_instrumented_executable_paths_map(\
                                            (TestCriteria.STRONG_MUTATION,))
        sm_map = list(crit2file[TestCriteria.STRONG_MUTATION].items())
        ERROR_HANDLER.assert_true(len(sm_map) == 1, \
                                "Expects exactly 1 file for Strong Mutation")
        exe_name, filepath = sm_map[0]
        return filepath + '.bc'
    #~ def get_test_gen_metamutant_bc()
#~ class CriteriaToolMart




from __future__ import print_function

import muteria.common.mix as common_mix

class CriteriaToolType(common_mix.EnumAutoName):
    USE_ONLY_CODE = "StaticCriteriaTool"
    #USE_CODE_AND_TESTS = "DynamicCriteriaTool"

    def get_tool_type_classname(self):
        return self.get_field_value()
    #~ def get_tool_type_classname():
#~ class CriteriaToolType

class TestCriteria(common_mix.EnumAutoName):
    STATEMENT_COVERAGE = "statement_coverage"
    BRANCH_COVERAGE = "branch_coverage"
    FUNCTION_COVERAGE = "function_coverage"

    MUTANT_COVERAGE = "mutant_coverage"
    WEAK_MUTATION = "weak_mutation"
    STRONG_MUTATION = "strong_mutation"
#~ class TestCriteria

CRITERIA_SEQUENCE = [
    {
        TestCriteria.STATEMENT_COVERAGE, TestCriteria.BRANCH_COVERAGE, TestCriteria.FUNCTION_COVERAGE,
    },
    {
        TestCriteria.MUTANT_COVERAGE,
    },
    {
        TestCriteria.WEAK_MUTATION,
    },
    {
        TestCriteria.STRONG_MUTATION,
    },
]
"""
"""

CRITERIA_TOOL_TYPES_SCHEDULING = [(CriteriaToolType.USE_ONLY_CODE,)]
"""
"""

CRITERIA_REQUIRING_OUTDIFF_WITH_PROGRAM = [TestCriteria.STRONG_MUTATION,]
"""
"""


from muteria.drivers.testgeneration.tools_by_languages.c.semu.semu \
                                                    import TestcasesToolSemu 
StaticTestcaseTool = None 
DynamicTestcaseTool = TestcasesToolSemu


""" TODO: call coverage
"""

from __future__ import print_function

import os
import sys
import re
import shutil
import shlex
import logging
import subprocess

import muteria.common.mix as common_mix
import muteria.common.fs as common_fs

from muteria.repositoryandcode.codes_convert_support import CodeFormats
from muteria.repositoryandcode.callback_object import DefaultCallbackObject

from muteria.drivers.criteria.base_testcriteriatool import BaseCriteriaTool
from muteria.drivers.criteria import TestCriteria
from muteria.drivers import DriversUtils

ERROR_HANDLER = common_mix.ErrorHandler

class CriteriaToolGCov(BaseCriteriaTool):
    def __init__(self, *args, **kwargs):
        BaseCriteriaTool.__init__(self, *args, **kwargs)
        self.instrumentation_details = os.path.join(\
                    self.instrumented_code_storage_dir, '.instru.meta.json')
        self.gcov_files_list_filename = "gcov_files.json"
        self.gc_files_dir = os.path.join(\
                                        self.criteria_working_dir, "gcno_gcda")
        self.gcov_gdb_wrapper_template = os.path.join(\
                                                os.path.dirname(__file__), \
                                                'gcov-gdb_call_wrapper.sh.in')
        self.gcov_gdb_wrapper_sh = os.path.join(\
                                        self.instrumented_code_storage_dir, \
                                                    "tmp_gcov_gdb_wrapper.sh")

        # clean any possible gcda file
        for file_ in self._get_gcda_list():
            os.remove(file_)
        for file_ in self._get_gcov_list():
            os.remove(file_)
    #~ def __init__()

    def _get_gcov_list(self):
        gcov_files = []
        for root, dirs, files in os.walk(self.gc_files_dir):
            for file_ in files:
                if file_.endswith('.gcov'):
                    gcov_files.append(os.path.join(root, file_))
        return gcov_files
    #~ def _get_gcov_list()

    def _get_gcda_list(self):
        gcda_files = []
        for root, dirs, files in os.walk(self.gc_files_dir):
            for file_ in files:
                if file_.endswith('.gcda'):
                    gcda_files.append(os.path.join(root, file_))
        return gcda_files
    #~ def _get_gcda_list()

    class InstrumentCallbackObject(DefaultCallbackObject):
        def after_command(self):
            if self.op_retval != common_mix.GlobalConstants.COMMAND_SUCCESS:
                ERROR_HANDLER.error_exit("Build failed", __file__)
            gc_files_dir, rel_path_map = self.post_callback_args
            for _, obj in list(self.source_files_to_objects.items()):
                rel_raw_filename, _ = os.path.splitext(obj)
                gcno_file = rel_raw_filename + '.gcno' 
                relloc = os.path.join(gc_files_dir, os.path.dirname(obj))
                if not os.path.isdir(relloc):
                    os.makedirs(relloc)
                abs_gcno = os.path.join(self.repository_rootdir, gcno_file)
                ERROR_HANDLER.assert_true(os.path.isfile(abs_gcno), \
                                    "gcno file missing after build", __file__)
                shutil.copy2(abs_gcno, os.path.join(gc_files_dir, gcno_file))
            self._copy_from_repo(rel_path_map)
            return DefaultCallbackObject.after_command(self)
        #~ def after_command()
    #~ class InstrumentCallbackObject

    @classmethod
    def installed(cls, custom_binary_dir=None):
        """ Check that the tool is installed
            :return: bool reprenting whether the tool is installed or not 
                    (executable accessible on the path)
                    - True: the tool is installed and works
                    - False: the tool is not installed or do not work
        """
        for prog in ('gcc', 'gcov'):
            if custom_binary_dir is not None:
                prog = os.path.join(custom_binary_dir, prog)
            if not DriversUtils.check_tool(prog=prog, args_list=['--version'],\
                                                    expected_exit_codes=[0]):
                return False
        return True
    #~ def installed()

    @classmethod
    def _get_meta_instrumentation_criteria(cls):
        """ Criteria where all elements are instrumented in same file
            :return: list of citeria
        """
        return [
                TestCriteria.STATEMENT_COVERAGE,
                TestCriteria.BRANCH_COVERAGE,
                TestCriteria.FUNCTION_COVERAGE,
               ]
    #~ def _get_meta_instrumentation_criteria()

    @classmethod
    def _get_separated_instrumentation_criteria(cls):
        """ Criteria where all elements are instrumented in different files
            :return: list of citeria
        """
        return []
    #~ def _get_separated_instrumentation_criteria()

    def get_instrumented_executable_paths_map(self, enabled_criteria):
        # use wrapper if gdb is installed
        using_gdb_wrapper = DriversUtils.check_tool(prog='gdb', \
                                                    args_list=['--version'], \
                                                    expected_exit_codes=[0])
        if using_gdb_wrapper:
            # XXX: Docker has issues running programs in Docker, 
            # unless the container is ran with the following arguments:
            #
            # docker run --cap-add=SYS_PTRACE \
            #               --security-opt seccomp=unconfined ...
            #
            # We check that it is fine by testing on echo
            ret, _, _ = DriversUtils.execute_and_get_retcode_out_err(
                                        prog='gdb', \
                                        args_list=['--batch-silent', \
                                                    '--quiet', 
                                                    '--return-child-result', 
                                                    '-ex', '"run"', 
                                                    '--args', 'echo'], \
                                        out_on=False, err_on=False)
            using_gdb_wrapper = (ret == 0)

        crit_to_exes_map = {}
        obj = common_fs.loadJSON(self.instrumentation_details)
        #exes = [p for _, p in list(obj.items())]
        for name in obj:
            obj[name] = os.path.join(self.instrumented_code_storage_dir, \
                                                                    obj[name])
            
        if using_gdb_wrapper:
            # Using GDB WRAPPER
            with open(self.gcov_gdb_wrapper_template) as f:
                template_str = f.read()
            single_exe = obj[list(obj)[0]]
            template_str = template_str.replace(\
                                'MUTERIA_GCOV_PROGRAMEXE_PATHNAME', single_exe)
            with open(self.gcov_gdb_wrapper_sh, 'w') as f:
                f.write(template_str)
            shutil.copymode(single_exe, self.gcov_gdb_wrapper_sh)
            obj[list(obj)[0]] = self.gcov_gdb_wrapper_sh

        exes_map = obj
        for criterion in enabled_criteria:
            crit_to_exes_map[criterion] = exes_map
        return crit_to_exes_map
    #~ def get_instrumented_executable_paths_map()

    def get_criterion_info_object(self, criterion):
        return None
    #~ def get_criterion_info_object(self, criterion)

    def _get_criterion_element_executable_path(self, criterion, element_id):
        ERROR_HANDLER.error_exit("not applicable for gcov", __file__)
    #~ def _get_criterion_element_executable_path

    def _get_criterion_element_environment_vars(self, criterion, element_id):
        '''
            return: python dictionary with environment variable as key
                     and their values as value (all strings)
        '''
        ERROR_HANDLER.error_exit("not applicable for gcov", __file__)
    #~ def _get_criterion_element_environment_vars()

    def _get_criteria_environment_vars(self, result_dir_tmp, enabled_criteria):
        '''
        return: python dictionary with environment variable as key
                     and their values as value (all strings)
        '''
        return {e:None for e in enabled_criteria}
    #~ def _get_criteria_environment_vars()

    def _collect_temporary_coverage_data(self, criteria_name_list, \
                                            test_execution_verdict, \
                                            used_environment_vars, \
                                            result_dir_tmp, \
                                            testcase):
        ''' get gcov files from gcda files into result_dir_tmp
        '''
        prog = 'gcov'
        if self.custom_binary_dir is not None:
            prog = os.path.join(self.custom_binary_dir, prog)
            ERROR_HANDLER.assert_true(os.path.isfile(prog), \
                            "The tool {} is missing from the specified dir {}"\
                                        .format(os.path.basename(prog), \
                                            self.custom_binary_dir), __file__)

        cov2flags = {
                    TestCriteria.STATEMENT_COVERAGE: [],
                    TestCriteria.BRANCH_COVERAGE: ['-b', '-c'],
                    TestCriteria.FUNCTION_COVERAGE: ['-f'],
                }

        args_list = []
        for criterion in criteria_name_list:
            args_list += cov2flags[criterion]

        gcda_files = self._get_gcda_list()

        raw_filename_list = [os.path.splitext(f)[0] for f in gcda_files]
        args_list += raw_filename_list
        
        if len(gcda_files) > 0:
            # TODO: When gcov generate coverage for different files with
            # same name filename but located at diferent dir. Avoid override.
            # Go where the gcov will be looked for
            cwd = os.getcwd()
            os.chdir(self.gc_files_dir)

            # collect gcda (gcno)
            r, _, _ = DriversUtils.execute_and_get_retcode_out_err(prog=prog, \
                                        args_list=args_list, out_on=False, \
                                                                err_on=False)

            os.chdir(cwd)
            
            if r != 0:
                ERROR_HANDLER.error_exit("Program {} {}.".format(prog,\
                        'error collecting coverage is problematic'), __file__)
            
            # delete gcda
            for gcda_f in gcda_files:
                os.remove(gcda_f)
            
        common_fs.dumpJSON(self._get_gcov_list(), \
                                os.path.join(result_dir_tmp,\
                                                self.gcov_files_list_filename))
        #else:
        #    ERROR_HANDLER.error_exit(\
        #                "Testcase '{}' did not generate gcda".format(testcase))
    #~ def _collect_temporary_coverage_data()

    def _extract_coverage_data_of_a_test(self, enabled_criteria, \
                                    test_execution_verdict, result_dir_tmp):
        ''' read json files and extract data
            return: the dict of criteria with covering count
            # TODO: Restrict to returning coverage of specified headers files
        '''
        gcov_list = common_fs.loadJSON(os.path.join(result_dir_tmp,\
                                                self.gcov_files_list_filename))
        
        res = {c: {} for c in enabled_criteria}

        func_cov = None
        branch_cov = None
        statement_cov = {}
        if TestCriteria.FUNCTION_COVERAGE in enabled_criteria:
            func_cov = res[TestCriteria.FUNCTION_COVERAGE]
        if TestCriteria.BRANCH_COVERAGE in enabled_criteria:
            branch_cov = res[TestCriteria.BRANCH_COVERAGE]
        if TestCriteria.STATEMENT_COVERAGE in enabled_criteria:
            statement_cov = res[TestCriteria.STATEMENT_COVERAGE]

        # Sources of interest
        _, src_map = self.code_builds_factory.repository_manager.\
                                                    get_relative_exe_path_map()

        for gcov_file in gcov_list:
            with open(gcov_file) as fp:
                last_line = None
                src_file = None
                for raw_line in fp:
                    line = raw_line.strip()
                    col_split = [v.strip() for v in line.split(':')]

                    if len(col_split) > 2 and col_split[1] == '0':
                        # preamble
                        if col_split[2] == "Source":
                            src_file = col_split[3]
                            if src_file not in src_map:
                                # src not in considered
                                break
                    elif line.startswith("function "):
                        # match function
                        parts = line.split()
                        ident = DriversUtils.make_meta_element(parts[1], \
                                                                    src_file)
                        func_cov[ident] = int(parts[3])
                    elif line.startswith("branch "):
                        # match branch
                        parts = line.split()
                        ident = DriversUtils.make_meta_element(parts[1], \
                                                                    last_line)
                        if parts[2:4] == ['never', 'executed']:
                            branch_cov[ident] = 0
                        else:
                            branch_cov[ident] = int(parts[3])

                    elif len(col_split) > 2 and \
                                            re.match(r"^\d+$", col_split[1]):
                        # match line
                        if col_split[0] == '-':
                            continue
                        last_line = DriversUtils.make_meta_element(\
                                                        col_split[1], src_file)
                        if col_split[0] in ('#####', '====='):
                            exec_count = 0
                        else:
                            exec_count = \
                                    int(re.findall(r'^\d+', col_split[0])[0])
                        statement_cov[last_line] = exec_count

        # delete gcov files
        for gcov_f in self._get_gcov_list():
            os.remove(gcov_f)

        return res
    #~ def _extract_coverage_data_of_a_test()

    def _do_instrument_code (self, exe_path_map, \
                                        code_builds_factory, \
                                        enabled_criteria, parallel_count=1):
        # Setup
        if os.path.isdir(self.instrumented_code_storage_dir):
            shutil.rmtree(self.instrumented_code_storage_dir)
        os.mkdir(self.instrumented_code_storage_dir)
        if os.path.isdir(self.gc_files_dir):
            try:
                shutil.rmtree(self.gc_files_dir)
            except PermissionError:
                self._dir_chmod777(self.gc_files_dir)
                shutil.rmtree(self.gc_files_dir)
        os.mkdir(self.gc_files_dir, mode=0o777)

        prog = 'gcc'
        if self.custom_binary_dir is not None:
            prog = os.path.join(self.custom_binary_dir, prog)
            ERROR_HANDLER.assert_true(os.path.isfile(prog), \
                            "The tool {} is missing from the specified dir {}"\
                                        .format(os.path.basename(prog), \
                                            self.custom_binary_dir), __file__)

        flags = ['-g', '--coverage', '-fprofile-dir='+self.gc_files_dir, '-O0']
        additionals = ["-fkeep-inline-functions"]
        
        # get gcc version
        ret, out, err = DriversUtils.execute_and_get_retcode_out_err(\
                                                        prog, ['-dumpversion'])
        ERROR_HANDLER.assert_true(ret == 0, "'gcc -dumpversion' failed'")
        
        # if version > 6.5
        if int(out.split('.')[0]) >= 6:
            if int(out.split('.')[0]) > 6 or int(out.split('.')[1]) > 5:
                additionals += ["-fkeep-static-functions"]
        
        flags += additionals
        
        rel_rel_path_map = {}
        rel_abs_path_map = {}
        exes, _ = code_builds_factory.repository_manager.\
                                                    get_relative_exe_path_map()
        ERROR_HANDLER.assert_true(len(exes) == 1, \
                                "Only single exe supported in GCOV", __file__)
        for exe in exes:
            filename = os.path.basename(exe)
            rel_rel_path_map[exe] = filename
            rel_abs_path_map[exe] = os.path.join(\
                                self.instrumented_code_storage_dir, filename)

        self.instrument_callback_obj = self.InstrumentCallbackObject()
        self.instrument_callback_obj.set_post_callback_args(\
                                        (self.gc_files_dir, rel_abs_path_map))
        pre_ret, ret, post_ret = code_builds_factory.transform_src_into_dest(\
                        src_fmt=CodeFormats.C_SOURCE,\
                        dest_fmt=CodeFormats.NATIVE_CODE,\
                        src_dest_files_paths_map=None,\
                        compiler=prog, flags_list=flags, clean_tmp=True, \
                        reconfigure=True, \
                        callback_object=self.instrument_callback_obj)
        
        # Check
        if ret == common_mix.GlobalConstants.COMMAND_FAILURE:
            ERROR_HANDLER.error_exit("Program {} {}.".format(prog,\
                                        'built problematic'), __file__)

        # write down the rel_path_map
        ERROR_HANDLER.assert_true(not os.path.isfile(\
                self.instrumentation_details), "must not exist here", __file__)
        common_fs.dumpJSON(rel_rel_path_map, self.instrumentation_details)

    #~ def _do_instrument_code()
#~ class CriteriaToolGCov



"""
 TODO: fully implement this
"""
from __future__ import print_function

import muteria.common.fs as common_fs
import muteria.common.mix as common_mix

from muteria.drivers.criteria import TestCriteria

ERROR_HANDLER = common_mix.ErrorHandler

class CriterionElementInfoObject(object):
    DATA_KEY = "DATA"
    SUMMARY_KEY = "SUMMARY"
    CUSTOM_KEY = "CUSTOM"
    def __init__(self):
        self.data = {
            self.DATA_KEY: {},
            self.SUMMARY_KEY: None,
            self.CUSTOM_KEY: None,
        }
    #~ def __init__()

    def load_from_file(self, file_path):
        self.data = common_fs.loadJSON(file_path)
    #~ def load_from_file()

    def write_to_file(self, file_path):
        common_fs.dumpJSON(self.data, file_path, pretty=True)
    #~ def write_to_file()

    def add_element (self, element_name, **kwargs):
        ERROR_HANDLER.assert_true(not self.has_element(element_name), \
                            "Test is already present in this: {}".format(\
                                                    element_name), __file__)
        self.data[self.DATA_KEY][element_name] = kwargs
    #~def add_element ():

    def get_element_data(self, element_name):
        ERROR_HANDLER.assert_true(self.has_element(element_name), \
                            "Test is not present in this: {}".format(\
                                                    element_name), __file__)
        return self.data[self.DATA_KEY][element_name]
    #~ def get_element_data()

    def has_element(self, element_name):
        return element_name in self.data[self.DATA_KEY]
    #~ def has_element():

    def get_elements_list(self):
        return list(self.data[self.DATA_KEY].keys())
    #~ def get_elements_list()

    def update_using(self, toolname, old2new_elements, old_element_info_obj):
        # Update elements
        for old_elem_name, new_elem_name in list(old2new_elements.items()):
            ERROR_HANDLER.assert_true(\
                        old_element_info_obj.has_element(old_elem_name), \
                        "TO not present in old_element_info_obj: {}".format(\
                                                    old_elem_name), __file__)
            
            ERROR_HANDLER.assert_true(not self.has_element(new_elem_name), \
                            "TO is already present in this: {}".format(\
                                                    new_elem_name), __file__)
            self.data[self.DATA_KEY][new_elem_name] = \
                        old_element_info_obj.data[self.DATA_KEY][old_elem_name]
        # Update Summary
        if self.data[self.SUMMARY_KEY] is None:
            self.data[self.SUMMARY_KEY] = {}
        self.data[self.SUMMARY_KEY][toolname] = \
                        old_element_info_obj.data[self.SUMMARY_KEY]

        # Update Custom
        if self.data[self.CUSTOM_KEY] is None:
            self.data[self.CUSTOM_KEY] = {}
        self.data[self.CUSTOM_KEY][toolname] = \
                        old_element_info_obj.data[self.CUSTOM_KEY]
    #~ def update_using()

    def remove_element(self, element_name):
        ERROR_HANDLER.assert_true(self.has_element(element_name), \
                    "Removing an unexisting element: {}".format(element_name),\
                                                                    __file__)
        del self.data[self.DATA_KEY][element_name]
    #~ def remove_element()

    def get_summary(self):
        return self.data[self.SUMMARY_KEY]
    #~ def get_summary():

    def get_custom(self):
        return self.data[self.CUSTOM_KEY]
    #~ def get_custom():
#~ class CriterionElementInfoObject(object):

class MutantsInfoObject(CriterionElementInfoObject):
    def __init__(self):
        CriterionElementInfoObject.__init__(self)
    #~ def __init__()

    def add_element (self, element_name, mutant_type=None, 
                        mutant_locs=None, mutant_function_name=None, **kwargs):
        CriterionElementInfoObject.add_element(self, element_name, \
                                    mutant_type=mutant_type, \
                                    mutant_srclocs=mutant_locs, \
                                    mutant_function_name=mutant_function_name,\
                                    **kwargs)
    #~def add_element ():

#~ class MutantsInfoObject(object):

"""
NOTE: For new citeria,
Update this dict with the criterion Info module
"""
CriteriaToInfoObject = {
    TestCriteria.STRONG_MUTATION: MutantsInfoObject,
    TestCriteria.WEAK_MUTATION: MutantsInfoObject,
    TestCriteria.MUTANT_COVERAGE: MutantsInfoObject,
}



"""
This module implement the main controller class, which is the entry point
of the execution.
The entry class is `MainController` and the entrypoint method is 
`MainController.mainrun`

TODO:
    0. write the out directory structure component properly 
        (This is shared between the 3 modes: RUN, REVERT and NAVIGATE)

    Pipeline:
    1. load the default configuration.
    2. parse the command lines, then load the users config, then update the
        default configs.
    3. setup the ERROR_HANDLER module (pass the repodir).
    4. setup the log facility.
    5. based on the loaded config and command mode, call the relevant 
        mode executor with its configs. (RUN, REVERT, NAVIGATE)

    a) RUN will implement checkpoint and run tasks....
"""

from __future__ import print_function
import os, sys
import glob
import copy

import importlib
import logging

import muteria.common.mix as common_mix
import muteria.common.fs as common_fs
import muteria.common.matrices as common_matrices

import muteria.configmanager.configurations as configurations
from muteria.configmanager.helper import ConfigurationHelper
from muteria.repositoryandcode.repository_manager import RepositoryManager

# from this package
import muteria.controller.logging_setup as logging_setup
import muteria.controller.explorer as explorer
import muteria.controller.executor as executor

ERROR_HANDLER = common_mix.ErrorHandler

class MainController (object):
    '''
    This class implements the main controlle which will organize the execution
    and reporting.

    :param execution_config: Configuration data for execution setting (such as:
                            execution order, what to execute, ...)
    :param reporting_config: Configuration for reporting (coverage, 
                            mutation score, ...)
    :param tools_config: Configurations of the tools involved (test tool, 
                            mutation tool, code coverage tools)
    :param project_config: Configurations of the project to analyze
    :param output_pathdir: Path to the directory where the execution take place
                            and the resulting data are stored.
    '''
    def __init__(self): 
        # Set the logger temporarily (before the log file can be dispo)
        logging_setup.console_tmp_log_setup()
    #~ def __init__()

    def internal_infos(self, config):
        # TODO
        ERROR_HANDLER.error_exit("FIXME: TODO: Implement the internal")
    #~ def internal_infos()

    def view(self, top_timeline_explorer, config):
        '''
        Method used to navigate in the output dir and make simple queries
        '''
        # TODO
        ERROR_HANDLER.error_exit("FIXME: TODO: Implement the View")
    #~ def view()

    #def log_run_summary(self):
    #    '''
    #    Log information showing a new run of resume of previous run
    #    '''

    def raw_config_main(self, raw_config):
        """
        TODO: Deal with config change at different runs
                Notify the user that the config changed and show the diff
        TODO: add option to load the config from prev if not specified()
        """
        # Create config from raw config
        final_conf = ConfigurationHelper.get_finalconf_from_rawconf(\
                                                        raw_conf=raw_config)

        # Call main with the result of raw config
        self.main(final_config=final_conf)
    #~ def raw_config_main()

    def main(self, final_config):
        '''
        Entry point function using the final configuration object
        '''
        # XXX Create TopExplorer
        top_timeline_explorer = explorer.TopExplorer(\
                                        final_config.OUTPUT_ROOT_DIR.get_val())

        # XXX Actual Execution based on the mode
        mode = final_config.RUN_MODE.get_val()

        if mode == configurations.SessionMode.EXECUTE_MODE:
            # Executor
            # XXX Main Execution
            exec_obj = executor.Executor(final_config, top_timeline_explorer)
            exec_obj.main()
        if mode == configurations.SessionMode.CUSTOM_EXECUTION_MODE:
            # Custom execution into other folder
            exec_obj = executor.Executor(final_config, top_timeline_explorer)
            exec_obj.custom_execution()
        elif mode == configurations.SessionMode.RESTORE_REPOS_MODE:
            # Restore the project repo dir's files that could have been changed.
            # This do not remove the possibly added files or folders
            as_initial = final_config.restore.AS_INITIAL.get_val()
            repo_mgr = executor.Executor.create_repo_manager(final_config)
            repo_mgr.revert_repository(as_initial=as_initial)
        elif mode == configurations.SessionMode.VIEW_MODE:
            # View
            self.view(top_timeline_explorer, final_config)
        elif mode == configurations.SessionMode.INTERNAL_MODE:
            # Internal Helpers
            self.internal_infos(final_config)
        logging.info("All Done!")
    #~ def main()
#~ class MainController



#
# [LICENCE]
#
""" Default criteria test execution optimizer class
"""

from __future__ import print_function
import os
import sys
import copy
import logging

import muteria.common.mix as common_mix

from muteria.drivers.optimizers.criteriatestexecution.\
                                base_criteria_test_execution_optimizer \
                                    import BaseCriteriaTestExecutionOptimizer

from muteria.drivers.optimizers.testexecution.tools.default \
                                                import TestExecutionOptimizer

ERROR_HANDLER = common_mix.ErrorHandler

class CriteriaTestExecutionOptimizer(BaseCriteriaTestExecutionOptimizer):

    #######################################################################
    ##################### Methods implemented ############################
    #######################################################################

    @classmethod
    def installed(cls, custom_binary_dir=None):
        """ Check that the tool is installed
            :return: bool reprenting whether the tool is installed or not 
                    (executable accessible on the path)
                    - True: the tool is installed and works
                    - False: the tool is not installed or do not work
        """
        return True
    #~ def installed()

    def reset (self, toolalias, test_objective_list, test_list, **kwargs):
        """ Reset the optimizer
        """
        self.test_objective_ordered_list = copy.deepcopy(test_objective_list)
        self.pointer = 0
        self.test_objective_to_test_execution_optimizer = {
            to: TestExecutionOptimizer(self.config, self.explorer) \
            for to in self.test_objective_ordered_list
        }
        for _, teo in self.test_objective_to_test_execution_optimizer.items():
            teo.reset(None, test_list, disable_reset=True)
    #~ def reset()
#~ class CriteriaTestExecutionOptimizer




"""
    Note: The raw configuration is a configuration that has the exact
    same representation as in the file but loaded in memory as a dict. 
    (Example: module given in file still have the filename or function
            given as string to be evaluated are also still string).
    But the final conf has everything loaded and ready to use
"""

from __future__ import print_function

import os
import sys
import argparse
import logging
import importlib
import copy

import muteria.common.mix as common_mix

import muteria.drivers.criteria as criteria
import muteria.drivers.testgeneration as testgeneration

import muteria.drivers.optimizers.criteriatestexecution.optimizerdefs as \
                                                                crit_opt_module
from muteria.drivers.optimizers.criteriatestexecution\
                                    .base_criteria_test_execution_optimizer \
                                    import BaseCriteriaTestExecutionOptimizer

import muteria.configmanager.configurations as configurations
from muteria.configmanager.configurations import CompleteConfiguration
from muteria.configmanager.configurations import ConfigElement

import muteria.controller.checkpoint_tasks as checkpoint_tasks

ERROR_HANDLER = common_mix.ErrorHandler

class ConfigurationHelper(object):

    @classmethod
    def _get_available_default_raw_conf_by_language(cls, languages):
        res = {}
        for lang in languages:
            res[lang] = cls._get_default_raw_conf(lang)
        return res
    #~ def _get_available_default_raw_conf_by_language()

    @classmethod
    def _get_default_raw_conf(cls, language):
        config_default = "muteria.configmanager.defaults"
        # Get common default conf
        com_default_conf = cls._load_raw_conf_from_file(\
                    module_str=".".join([config_default, 'common_defaults']),\
                    info="common config module", \
                    must_exist=True)
        ERROR_HANDLER.assert_true(com_default_conf is not None, \
                                                'invalid for comm', __file__)
        
        # Get language specific default conf
        lang_default_conf = cls._load_raw_conf_from_file(\
                module_str=".".join([config_default, 'languages', language]),\
                info="language "+language+" config module",\
                must_exist=False)

        # create an object that is an update of common_default_conf by 
        # lang_default_conf
        com_conf_dict = cls._get_object_params_vals_as_dict(com_default_conf)
        if lang_default_conf is None:
            lang_conf_dict = {}
        else:
            lang_conf_dict = \
                        cls._get_object_params_vals_as_dict(lang_default_conf)

        ## Make sure default has all parameters
        c_tmp = cls._get_object_params_vals_as_dict(CompleteConfiguration)
        useless = set(com_conf_dict) - set(c_tmp)
        ERROR_HANDLER.assert_true(\
                        len(useless) + len(c_tmp) == len(com_conf_dict),\
                        "Some Configs params not in common default", __file__)

        # get result
        ## remove irrelevant
        for u in useless:
            del com_conf_dict[u]
        ## update
        res = cls._get_update_left_with_right_raw_conf(com_conf_dict,\
                                                lang_conf_dict, same_key=False)
        return res
    #~ def _get_default_raw_conf()

    @staticmethod
    def _get_object_params_vals_as_dict(obj):
        """ Get all fields of obj that are not hidden
        """
        res = {}
        for param, val in list(vars(obj).items()):
            if param.startswith('_'):
                continue
            res[param] = val
        return res
    #~ def _get_object_params_vals_as_dict()

    @staticmethod
    def _load_raw_conf_from_file(module_str, info="module", must_exist=True):
        res = None
        try:
            res = importlib.import_module(module_str)
        except ImportError as ie:
            if must_exist:
                ERROR_HANDLER.error_exit("{} {} {}. \n {}.".format(\
			     "Failed to load", info, module_str, str(ie)), \
                                                                   __file__)
        except SyntaxError as se:
            ERROR_HANDLER.error_exit("{} {} {}. \n {}.".format(\
                            "Syntax error in", info, module_str, str(se)), \
                                                                   __file__)
        return res
    #~ def _load_raw_conf_from_file()

    @classmethod
    def _get_update_rawconf_with_file(cls, raw_conf, raw_conf_filename):
        """
        load config from file and update raw conf with its contents
        """
        path, filename = os.path.split(os.path.normpath(\
                                        os.path.abspath(raw_conf_filename)))
        if path:
            sys.path.insert(0, path)
        ERROR_HANDLER.assert_true(filename.endswith('.py'), "{}{}{}".format(\
                                    "invalid conf file (", filename, \
                                    "). must be python source file"), __file__)
        mod_name = filename[:-len('.py')]
        fconf = cls._load_raw_conf_from_file(mod_name, info="conf file", \
                                                            must_exist=True)
        if path:
            ERROR_HANDLER.assert_true(sys.path[0] == path, "BUG", __file__)
            sys.path.pop(0)
        if fconf is None:
            fconf_dict = {}
        else:
            fconf_dict = cls._get_object_params_vals_as_dict(fconf)

        # update
        res = cls._get_update_left_with_right_raw_conf(raw_conf, fconf_dict,\
                                                                same_key=False)
        return res
    #~ def _get_update_rawconf_with_file()

    @staticmethod
    def _get_update_left_with_right_raw_conf(left, right, same_key=True):
        """ Update the keys of left with the corresponding keys in right
            into a new dict and return it.
        """
        if same_key:
            ERROR_HANDLER.assert_true(set(left) == set(right), \
                                                    "different keys", __file__)
        res = dict(left)
        for k, v in list(right.items()):
            if k in res:
                res[k] = v
        return res
    #~ def _get_update_left_with_right_raw_conf()

    @classmethod
    def _make_conf_class_from_dict(cls, dict_obj):
        cc = CompleteConfiguration()

        if set(dict_obj) != set(cls._get_object_params_vals_as_dict(\
                                                    CompleteConfiguration)):
            in_obj_only = set(dict_obj) - \
                                set(cls._get_object_params_vals_as_dict(\
                                                        CompleteConfiguration))
            final_templ_only = set(cls._get_object_params_vals_as_dict(\
                                        CompleteConfiguration)) - set(dict_obj)
            ERROR_HANDLER.error_exit("config missmatch: {} {}. {} {}".format(\
                            "in_obj_only", in_obj_only,
                            "final template only", final_templ_only), __file__)
        for k, v in list(dict_obj.items()):
            setattr(cc, k, v)
        return cc
    #~ def _make_conf_class_from_dict()

    @staticmethod
    def _make_tool_conf_from_raw(raw_dict_conf, tool_type_enum, 
                                default_tool_type, target_tool_class):
        tuc = 'tool_user_custom'
        ttype = 'tooltype'
        c_on = 'criteria_on'

        if ttype in raw_dict_conf:
            if raw_dict_conf[ttype]:
                ERROR_HANDLER.assert_true(\
                            hasattr(tool_type_enum, raw_dict_conf[ttype]), \
                        "Invalid test tool type: "+raw_dict_conf[ttype], \
                                                                    __file__)
                raw_dict_conf[ttype] = \
                                getattr(tool_type_enum, raw_dict_conf[ttype])
            else:
                raw_dict_conf[ttype] = default_tool_type

        if c_on in raw_dict_conf:
            if raw_dict_conf[c_on]:
                raw_dict_conf[c_on] = [getattr(criteria.TestCriteria, c) \
                                                for c in raw_dict_conf[c_on]]
            else:
                raw_dict_conf[c_on] = None

        if tuc in raw_dict_conf:
            if raw_dict_conf[tuc]:
                raw_dict_conf[tuc] = \
                            configurations.ToolUserCustom(**raw_dict_conf[tuc])
            else:
                raw_dict_conf[tuc] = None
        
        return target_tool_class(**raw_dict_conf)
    #~ def _make_tool_conf_from_raw()

    
    @classmethod
    def get_extend_file_raw_conf(cls, raw_conf_file, language):
        """ Take a raw conf file that might not have all the parameters and 
            load, then add the missing with default value into a new 
            returned raw conf
        """
        def_conf = cls._get_default_raw_conf(language)
        res = cls._get_update_rawconf_with_file(def_conf, raw_conf_file)
        return res
    #~ def get_extend_file_raw_conf()

    @classmethod
    def get_extend_raw_conf(cls, raw_conf, language):
        """ Take a raw conf that might not have all the parameters and 
            add the missing with default value into a new returned raw conf
        """
        def_conf = cls._get_default_raw_conf(language)
        res = cls._get_update_left_with_right_raw_conf(def_conf, raw_conf, \
                                                                same_key=False)
        return res
    #~ def get_extend_raw_conf()
        
    @classmethod
    def get_finalconf_from_rawconf(cls, raw_conf):
        """ Transform a raw conf into final conf
            TODO: Verify config consistency
        """
        conf = cls._make_conf_class_from_dict(raw_conf)
        
        if conf.ENABLED_CRITERIA is None:
            conf.ENABLED_CRITERIA = \
                        copy.deepcopy(conf.CRITERIA_TOOLS_CONFIGS_BY_CRITERIA)
        tmp = []
        for c in conf.ENABLED_CRITERIA:
            if not isinstance(c, criteria.TestCriteria):
                ERROR_HANDLER.assert_true(\
                            criteria.TestCriteria.has_element_named(c), \
                            "invalid test criterion: "+c)
                c = criteria.TestCriteria[c] 
            tmp.append(c)
        conf.ENABLED_CRITERIA = tmp

        tmp = []
        for c in conf.CRITERIA_WITH_OUTPUT_SUMMARY:
            if not isinstance(c, criteria.TestCriteria):
                ERROR_HANDLER.assert_true(\
                            criteria.TestCriteria.has_element_named(c), \
                            "invalid test criterion in out sum: "+c, __file__)
                c = criteria.TestCriteria[c] 
            if c not in conf.ENABLED_CRITERIA:
                continue
            tmp.append(c)
        conf.CRITERIA_WITH_OUTPUT_SUMMARY = tmp


        if conf.CRITERIA_SEQUENCE is None:
            conf.CRITERIA_SEQUENCE = copy.deepcopy(criteria.CRITERIA_SEQUENCE)
        for pos, group in enumerate(conf.CRITERIA_SEQUENCE):
            tmp = set()
            for c in group:
                if not isinstance(c, criteria.TestCriteria):
                    ERROR_HANDLER.assert_true(\
                                criteria.TestCriteria.has_element_named(c), \
                                "invalid test criterion in seq: "+c, __file__)
                    c = criteria.TestCriteria[c] 
                if c not in conf.ENABLED_CRITERIA:
                    continue
                tmp.add(c)
            conf.CRITERIA_SEQUENCE[pos] = tmp

        
        if conf.CRITERIA_REQUIRING_OUTDIFF_WITH_PROGRAM is None:
            conf.CRITERIA_REQUIRING_OUTDIFF_WITH_PROGRAM = copy.deepcopy(\
                            criteria.CRITERIA_REQUIRING_OUTDIFF_WITH_PROGRAM)
        tmp = []
        for c in conf.CRITERIA_REQUIRING_OUTDIFF_WITH_PROGRAM:
            if not isinstance(c, criteria.TestCriteria):
                ERROR_HANDLER.assert_true(\
                            criteria.TestCriteria.has_element_named(c), \
                            "invalid test criterion in outdiff: "+c, __file__)
                c = criteria.TestCriteria[c] 
            if c not in conf.ENABLED_CRITERIA:
                continue
            tmp.append(c)
        conf.CRITERIA_REQUIRING_OUTDIFF_WITH_PROGRAM = tmp
        
        tmp = {}
        for c, sel_tech in conf.CRITERIA_ELEM_SELECTIONS.items():
            if not isinstance(c, criteria.TestCriteria):
                ERROR_HANDLER.assert_true(\
                            criteria.TestCriteria.has_element_named(c), \
                        "invalid test criterion in crit elem selection: "+c, \
                                                                    __file__)
                c = criteria.TestCriteria[c] 
            if c not in conf.ENABLED_CRITERIA:
                continue
            tmp[c] = sel_tech
        conf.CRITERIA_ELEM_SELECTIONS = tmp
        
        tmp = []
        for tc in conf.TESTCASE_TOOLS_CONFIGS:
            if not isinstance(tc, configurations.TestcaseToolsConfig):
                tc = cls._make_tool_conf_from_raw(tc, \
                            testgeneration.TestToolType,\
                            testgeneration.TEST_TOOL_TYPES_SCHEDULING[0][0],\
                            configurations.TestcaseToolsConfig)
            tmp.append(tc)
        conf.TESTCASE_TOOLS_CONFIGS = tmp

        tmp = {}
        for crit, cc_list in list(\
                            conf.CRITERIA_TOOLS_CONFIGS_BY_CRITERIA.items()):
            for cc in cc_list:
                if not isinstance(cc, configurations.CriteriaToolsConfig):
                    cc = cls._make_tool_conf_from_raw(cc, \
                            criteria.CriteriaToolType,\
                            criteria.CRITERIA_TOOL_TYPES_SCHEDULING[0][0],\
                            configurations.CriteriaToolsConfig)
            # TODO: Criteria enabled verification
            if crit not in tmp:
                tmp[crit] = []
            tmp[crit].append(cc)
        conf.CRITERIA_TOOLS_CONFIGS_BY_CRITERIA = tmp

        # criteria Optimizer
        tmp = {}
        for c, opt in conf.CRITERIA_EXECUTION_OPTIMIZERS.items():
            if not isinstance(c, criteria.TestCriteria):
                ERROR_HANDLER.assert_true(\
                            criteria.TestCriteria.has_element_named(c), \
                            "invalid test criterion in opt: "+c, __file__)
                c = criteria.TestCriteria[c] 
            if c not in conf.ENABLED_CRITERIA:
                continue
            # use base criteria opt to give option to set one own optimizer 
            # in conf
            if not isinstance(opt, crit_opt_module.CriteriaOptimizers) and \
                    not isinstance(opt, BaseCriteriaTestExecutionOptimizer):
                ERROR_HANDLER.assert_true(crit_opt_module.CriteriaOptimizers.\
                                                    has_element_named(opt), \
                                        "Invalid criterion Optimizer: "+opt)
                opt = crit_opt_module.CriteriaOptimizers[opt] 
            ERROR_HANDLER.assert_true(\
                            crit_opt_module.check_is_right_optimizer(c, opt), \
                                        "Wrong optimizer for test criterion")
            tmp[c] = opt
        # Make sure that all criteria have optimizers
        for c in conf.ENABLED_CRITERIA:
            if c not in tmp:
                tmp[c] = crit_opt_module.CriteriaOptimizers.NO_OPTIMIZATION
        conf.CRITERIA_EXECUTION_OPTIMIZERS = tmp


        tmp = []
        for ct in conf.RE_EXECUTE_FROM_CHECKPOINT_META_TASKS:
            if not isinstance(ct, checkpoint_tasks.Tasks):
                ERROR_HANDLER.assert_true(\
                            checkpoint_tasks.Tasks.has_element_named(ct), \
                            "invalid checkpoint task: "+ct)
                ct = checkpoint_tasks.Tasks[ct] 
            tmp.append(ct)
        conf.RE_EXECUTE_FROM_CHECKPOINT_META_TASKS = tmp

        # NEXT here
        #TODO: Add optimizers ....

        # make configelement for each
        for k in raw_conf:
            setattr(conf, k, ConfigElement(val=getattr(conf, k)))

        return conf
    #~ def get_finalconf_of_rawconf()

#~ class ConfigurationHelper()



#
# [LICENCE]
#
"""
This module is used through MetaTestcaseTool class
Which access the relevant testcase tools as specified

The tools are organized by programming language
For each language, there is a folder for each tool, 
named after the tool in lowercase

Each testcase tool package have the following in the __init__.py file:
>>> import <Module>.<class extending BaseTestcaseTool> as TestcaseTool
"""

from __future__ import print_function
import os
import sys
import glob
import shutil
import copy
import logging
import multiprocessing
import joblib

import muteria.common.fs as common_fs
import muteria.common.matrices as common_matrices
import muteria.common.mix as common_mix

from muteria.drivers import ToolsModulesLoader
from muteria.drivers import DriversUtils

from muteria.drivers.checkpoint_handler import CheckPointHandler

from muteria.drivers.testgeneration.testcases_info import TestcasesInfoObject
from muteria.drivers.testgeneration import TestToolType

from muteria.drivers.testgeneration.custom_dev_testcase.custom_dev_testcase \
                                                        import CustomTestcases

ERROR_HANDLER = common_mix.ErrorHandler

class MetaTestcaseTool(object):
    ''' 
    '''

    TOOL_OBJ_KEY = "tool_obj"
    TOOL_WORKDIR_KEY = "tool_working_dir"
    TOOL_TYPE_KEY = "tool_type"

    # The Fault execution matrix has a unique key represented by this string
    FAULT_MATRIX_KEY = "fault_revealed"
    PROGRAM_EXECOUTPUT_KEY = "program"

    @classmethod
    def get_toolnames_by_types_by_language(cls):
        """ get imformation about the plugged-in testcase tool drivers.
            :return: a dict having the form:
                    {
                        language: {
                            TestToolType: [
                                (toolname, is_installed?)
                            ]
                        }
                    }
        """
        modules_dict = ToolsModulesLoader.get_tools_modules( \
                                            ToolsModulesLoader.TESTCASES_TOOLS)
        res = {}
        for language in modules_dict:
            res[language] = {}
            for toolname in modules_dict[language]:
                for tooltype in TestToolType:
                    tooltype_name = tooltype.get_field_value()
                    try:
                        # the tooltype is returned by config.get_tool_type()
                        TestcaseTool = getattr(\
                                            modules_dict[language][toolname],\
                                                                tooltype_name)
                        if TestcaseTool is not None:
                            if tooltype_name not in res[language]:
                                res[language][tooltype_name] = []
                            res[language][tooltype_name].append(\
                                        (toolname, TestcaseTool.installed()))
                    except AttributeError:
                        ERROR_HANDLER.error_exit("{} {} {} {}".format( \
                                "(REPORT BUG) The test case tool of type", \
                                tooltype_name,\
                                "is not present for test tool", toolname), \
                                                                    __file__)
        
        # Add custom dev test
        for language in res:
            tooltype = TestToolType.USE_ONLY_CODE
            res[language][tooltype.get_tool_type_classname()] = \
                                    (CustomTestcases.CUSTOM_TEST_TOOLNAME, \
                                                CustomTestcases.installed())
        return res
    #~ def get_toolnames_by_types_by_language()

    def __init__(self, language, tests_working_dir, code_builds_factory,
                                test_tool_config_list, head_explorer, 
                                hash_outlog=True):

        """ Initialize a meta testcase tool object.
        :type language:
        :param language:

        :type tests_working_dir:
        :param tests_working_dir:

        :type code_builds_factory:
        :param code_builds_factory:

        :type test_tool_config_list:
        :param test_tool_config_list:

        :raises:

        :rtype:
        """
        # Set Constants
        self.modules_dict = ToolsModulesLoader.get_tools_modules( \
                                            ToolsModulesLoader.TESTCASES_TOOLS)

        # Set Direct Arguments Variables
        self.language = language.lower()
        self.tests_working_dir = tests_working_dir
        self.code_builds_factory = code_builds_factory
        self.test_tool_config_list = test_tool_config_list
        self.head_explorer = head_explorer
        self.hash_outlog = hash_outlog

        # Verify Direct Arguments Variables
        ERROR_HANDLER.assert_true(self.tests_working_dir is not None, \
                                    "Must specify tests_working_dir", __file__)
        ERROR_HANDLER.assert_true(len(self.test_tool_config_list) == \
                                len(set([c.get_tool_config_alias() for c in \
                                            self.test_tool_config_list])), \
                        "some tool configs appear multiple times", __file__)

        # Set Indirect Arguments Variables
        self.checkpoints_dir = os.path.join(self.tests_working_dir, \
                                                            "_checkpoints_")
        self.testcases_info_file = \
                os.path.join(self.tests_working_dir, "testcasesInfos.json")
        self.flakiness_workdir = \
                            os.path.join(self.tests_working_dir, "Flakiness")

        # Verify Indirect Arguments Variables

        # Initialize other Fields
        self.testcases_configured_tools = {}
        self.checkpointer = None 

        # Make Initialization Computation ()
        ## Create dirs
        if not os.path.isdir(self.tests_working_dir):
            self.clear_working_dir()

        ## Set the checkpointer
        self.checkpointer = common_fs.CheckpointState(\
                                                *self._get_checkpoint_files())
        
        self.custom_devtest_toolalias = None

        # Create the diffent tools
        for idx in range(len(test_tool_config_list)):
            config = test_tool_config_list[idx]
            toolname = config.get_tool_name()
            toolalias = config.get_tool_config_alias()
            tool_working_dir = self._get_test_tool_out_folder(toolalias)
            tool_checkpointer = common_fs.CheckpointState(\
                            *self._get_test_tool_checkpoint_files(toolalias))
            self.checkpointer.add_dep_checkpoint_state(tool_checkpointer)
            self.testcases_configured_tools[toolalias] = {
                self.TOOL_OBJ_KEY: self._create_testcase_tool(toolname, \
                                                    tool_working_dir, config, \
                                                    tool_checkpointer),
                self.TOOL_WORKDIR_KEY: tool_working_dir,
                self.TOOL_TYPE_KEY: config.get_tool_type()
            }
    #~ def __init__()

    def _create_testcase_tool(self, toolname, tool_working_dir, \
                                                    config, tool_checkpointer):
        '''
            
        '''
        # Check for custom testcase
        if toolname == CustomTestcases.CUSTOM_TEST_TOOLNAME:
            TestcaseTool = CustomTestcases
            ERROR_HANDLER.assert_true(self.custom_devtest_toolalias is None, \
                        "Must specify custom dev tests only once", __file__)
            self.custom_devtest_toolalias = config.get_tool_config_alias()
        else:
            ERROR_HANDLER.assert_true( \
                toolname in self.modules_dict[self.language],
                "Invalid toolname given: {}".format(toolname), __file__)
            try:
                # the tooltype is returned by config.get_tool_type()
                TestcaseTool = getattr(\
                                self.modules_dict[self.language][toolname],\
                                config.get_tool_type().get_field_value())
            except AttributeError:
                ERROR_HANDLER.error_exit("{} {} {} {}".format( \
                                    "(REPORT BUG) The test case tool of type",\
                                    config.get_tool_type().get_field_value(),\
                                    "is not present for test tool", toolname),\
                                                                    __file__)

            ERROR_HANDLER.assert_true(TestcaseTool is not None, \
                                    "The {} language's tool {} {} {}.".format(\
                                    self.language, toolname, \
                                    "does not implement the test tool type", \
                                        config.get_tool_type().get_str()), \
                                                                    __file__)

        testcase_tool = TestcaseTool(tool_working_dir, \
                                        self.code_builds_factory, config, \
                                                    self.head_explorer, \
                                                    tool_checkpointer, \
                                                    parent_meta_tool=self)
        return testcase_tool
    #~ def _create_testcase_tool()

    def check_tools_installed(self):
        non_installed = []
        for toolalias, tool_obj in self.testcases_configured_tools.items():
            if not tool_obj[self.TOOL_OBJ_KEY].tool_installed():
                non_installed.append(toolalias)
        if len(non_installed) > 0:
            ERROR_HANDLER.error_exit("{}: {}".format(\
                            "The following Testcase tools are not installed", \
                            str(non_installed)))
    #~ def check_tools_installed()

    def get_devtest_toolalias(self):
        return self.custom_devtest_toolalias
    #~ def get_devtest_toolalias()


    def clear_working_dir(self):
        if os.path.isdir(self.tests_working_dir):
            shutil.rmtree(self.tests_working_dir)
        os.mkdir(self.tests_working_dir)
        if os.path.isdir(self.checkpoints_dir):
            shutil.rmtree(self.checkpoints_dir)
        os.mkdir(self.checkpoints_dir)
        for _, tool_dat in list(self.testcases_configured_tools.items()):
            tool_dat[self.TOOL_OBJ_KEY].clear_working_dir()
    #~ def clear_working_dir()    


    def _get_default_exe_path_map(self):
        exes, _ = self.code_builds_factory.repository_manager\
                                                .get_relative_exe_path_map()
        return {v:None for v in exes}
    #~ def _get_default_exe_path_map()

    def execute_testcase (self, meta_testcase, exe_path_map, env_vars, \
                                        timeout=None, \
                                        use_recorded_timeout_times=None, \
                                        recalculate_execution_times=False, \
                                        with_output_summary=True, \
                                        hash_outlog=None):
        '''
        Execute a test case with the given executable and 
        say whether it failed

        :param meta_testcase: string name of the test cases to execute
        :param exe_path_map: string representing the file system path to 
                        the executable to execute with the test
        :param env_vars: dict of environment variables to set before
                        executing the test ({<variable>: <value>})
        :type hash_outlog: bool
        :hash_outlog: decide whether to hash the outlog or not
        :returns: pair of:
                - boolean failed verdict of the test 
                        (True if failed, False otherwise)
                - test execution output log hash data object or None
        '''
        
        if exe_path_map is None:
            exe_path_map = self._get_default_exe_path_map()

        if hash_outlog is None:
            hash_outlog = self.hash_outlog

        # Find which test tool's the testcase is, then execute
        ttoolalias, testcase = DriversUtils.reverse_meta_element(meta_testcase)
        ERROR_HANDLER.assert_true( \
                            ttoolalias in self.testcases_configured_tools, \
                            "Test tool {} not registered".format(ttoolalias), \
                                                                    __file__)
        ttool = self.testcases_configured_tools[ttoolalias][self.TOOL_OBJ_KEY]
        return ttool.execute_testcase(testcase, exe_path_map, env_vars, \
                            timeout=timeout, \
                            use_recorded_timeout_times=\
                                                use_recorded_timeout_times, \
                            recalculate_execution_times=\
                                                recalculate_execution_times, \
                            with_output_summary=with_output_summary, \
                            hash_outlog=hash_outlog)
    #~ def execute_testcase()

    def runtests(self, meta_testcases=None, exe_path_map=None, env_vars=None, \
                        stop_on_failure=False, \
                        per_test_timeout=None, \
                        use_recorded_timeout_times=None, \
                        recalculate_execution_times=False, \
                        fault_test_execution_matrix_file=None, \
                        fault_test_execution_execoutput_file=None, \
                        with_output_summary=True, \
                        hash_outlog=None, \
                        test_prioritization_module=None, \
                        parallel_test_count=1, \
                        parallel_test_scheduler=None, \
                        restart_checkpointer=False,
                        finish_destroy_checkpointer=True):
        '''
        Execute the list of test cases with the given executable and 
        say, for each test case, whether it failed

        :param meta_testcases: list of test cases to execute
        :param exe_path_map: string representing the file system path to 
                        the executable to execute with the tests
        :param env_vars: dict of environment variables to set before
                        executing each test ({<variable>: <value>})
        :param stop_on_failure: decide whether to stop the test 
                        execution once a test fails
        :param fault_test_execution_matrix_file: Optional matrix file 
                        to store the tests' pass fail execution data
        :param fault_test_execution_execoutput_file: Optional output log file 
                        to store the tests' execution actual output (hashed)
        :param with_output_summary: decide whether to return outlog hash 
        :type hash_outlog: bool
        :hash_outlog: decide whether to hash the outlog or not
        :param test_prioritization_module: Specify the test prioritization
                        module. 
                        (TODO: Implement support)
        :param parallel_test_count: Specify the number of parallel test
                        Execution. must be an integer >= 1 or None.
                        When None, the max possible value is used.
        :param parallel_test_scheduler: Specify the function that will
                        handle parallel test scheduling by tool, using
                        the test execution optimizer. 
                        (TODO: Implement support)

        :type restart_checkointer: bool
        :param restart_checkointer: Decide whether to discard checkpoint
                        and restart anew.

        :type finish_destroy_checkpointer: bool
        :param finish_destroy_checkpointer: Decide whether to automatically 
                        destroy the checkpointer when done or not
                        Useful is caller has a checkpointer to update. 

        :returns: dict of testcase and their failed verdict.
                 {<test case name>: <True if failed, False if passed,
                    UNCERTAIN_TEST_VERDICT if uncertain>}
                 If stop_on_failure is True, only return the tests that 
                 have been executed until the failure
        '''
        
        ERROR_HANDLER.assert_true(meta_testcases is not None, \
                                            "Must specify testcases", __file__)

        # FIXME: Make sure that the support are implemented for 
        # parallelism and test prioritization. Remove the code bellow 
        # once supported:
        ERROR_HANDLER.assert_true(test_prioritization_module is None, \
                        "Must implement test prioritization support here", \
                                                                    __file__)
        ERROR_HANDLER.assert_true(parallel_test_scheduler is None, \
                    "Must implement parallel tests execution support here", \
                                                                    __file__)
        #~FIXMEnd

        # Check arguments Validity
        if exe_path_map is None:
            exe_path_map = self._get_default_exe_path_map()

        if hash_outlog is None:
            hash_outlog = self.hash_outlog

        ERROR_HANDLER.assert_true(parallel_test_count is None \
                                        or parallel_test_count >= 1, \
                                "invalid parallel tests count ({})".format(\
                                                parallel_test_count), __file__)

        # @Checkpoint: create a checkpoint handler
        cp_func_name = "runtests"
        cp_task_id = 1
        checkpoint_handler = \
                CheckPointHandler(self.get_checkpoint_state_object())
        if restart_checkpointer:
            checkpoint_handler.restart()
        if checkpoint_handler.is_finished():
            logging.warning("%s %s" %("The function 'runtests' is finished", \
                "according to checkpoint, but called again. None returned"))
            if common_mix.confirm_execution("%s %s" % ( \
                                        "Function 'runtests' is already", \
                                        "finished, do you want to restart?")):
                checkpoint_handler.restart()
                logging.info("Restarting the finished 'runtests'")
            else:
                ERROR_HANDLER.error_exit(err_string="%s %s %s" % (\
                        "Execution halted. Cannot continue because no value", \
                        " can be returned. Check the results of the", \
                        "finished execution"), call_location=__file__)

        # @Checkpoint: Get the saved payload (data kapt for each tool)
        # pair list of testfailed verdict and execution output
        meta_test_failedverdicts_outlog = \
                                    checkpoint_handler.get_optional_payload()
        if meta_test_failedverdicts_outlog is None:
            meta_test_failedverdicts_outlog = [{}, {}]

        # Make sure the tests are unique
        ERROR_HANDLER.assert_true(len(meta_testcases) == \
                                                len(set(meta_testcases)), \
                                        "not all tests are unique", __file__)

        testcases_by_tool = {}
        for meta_testcase in meta_testcases:
            ttoolalias, testcase = \
                            DriversUtils.reverse_meta_element(meta_testcase)
            if ttoolalias not in testcases_by_tool:
                testcases_by_tool[ttoolalias] = []
            testcases_by_tool[ttoolalias].append(testcase)
            
        candidate_aliases = []
        for tpos, ttoolalias in enumerate(testcases_by_tool.keys()):
            # @Checkpoint: Check whether already executed
            if not checkpoint_handler.is_to_execute(func_name=cp_func_name, \
                                                taskid=cp_task_id, \
                                                tool=ttoolalias):
                continue
            candidate_aliases.append(ttoolalias)

        shared_loc = multiprocessing.RLock()

        parallel_test_count_by_tool = {ta: 1 for ta in candidate_aliases}

        # tool with parallel test exec
        # TODO: find way to pass parallel count here
        if parallel_test_count is None:
            parallel_test_count = min(10, multiprocessing.cpu_count())
        
        # use parallel
        sub_parallel_count = 0 if parallel_test_count is None else \
                        parallel_test_count - len(parallel_test_count_by_tool)
        if sub_parallel_count > 0:
            para_tools = [tt for tt in candidate_aliases if \
                                self.testcases_configured_tools[ttoolalias]\
                                [self.TOOL_OBJ_KEY].can_run_tests_in_parallel()
                        ]
            para_tools.sort(reverse=True, \
                                    key=lambda x: len(testcases_by_tool[x]))
            para_tools_n_tests = sum(\
                            [len(testcases_by_tool[tt]) for tt in para_tools])
            
            used = 0
            for tt in para_tools:
                quota = int(len(testcases_by_tool[tt]) * sub_parallel_count / \
                                                            para_tools_n_tests)
                parallel_test_count_by_tool[tt] += quota
                used += quota
            for tt in para_tools:
                if used == sub_parallel_count:
                    break
                parallel_test_count_by_tool[tt] += 1


        def tool_parallel_test_exec(ttoolalias):
            # Actual execution
            found_a_failure=False
            ttool = \
                self.testcases_configured_tools[ttoolalias][self.TOOL_OBJ_KEY]
            test_failed_verdicts, test_execoutput = ttool.runtests( \
                                testcases_by_tool[ttoolalias], \
                                exe_path_map, env_vars, \
                                stop_on_failure, \
                                per_test_timeout=per_test_timeout,
                                use_recorded_timeout_times=\
                                    use_recorded_timeout_times, \
                                recalculate_execution_times=\
                                    recalculate_execution_times, \
                                with_output_summary=\
                                            with_output_summary, \
                                hash_outlog=hash_outlog, \
                                parallel_count=\
                                    parallel_test_count_by_tool[ttoolalias])
            with shared_loc:
                for testcase in test_failed_verdicts:
                    meta_testcase =  DriversUtils.make_meta_element(\
                                                        testcase, ttoolalias)
                    meta_test_failedverdicts_outlog[0][meta_testcase] = \
                                                test_failed_verdicts[testcase]
                    meta_test_failedverdicts_outlog[1][meta_testcase] = \
                                                    test_execoutput[testcase]
                    if not found_a_failure \
                                and test_failed_verdicts[testcase] == \
                                common_mix.GlobalConstants.COMMAND_UNCERTAIN:
                        found_a_failure = True

                # @Checkpoint: Chekpointing
                checkpoint_handler.do_checkpoint(func_name=cp_func_name, \
                                taskid=cp_task_id, \
                                tool=ttoolalias, \
                                opt_payload=meta_test_failedverdicts_outlog)
            return found_a_failure
        #~ def tool_parallel_test_exec()

        # minimum number of tests for parallelism
        ptest_tresh = 5

        if len(candidate_aliases) > 1 and len(meta_testcases) >= ptest_tresh \
                                        and parallel_test_count is not None \
                                        and parallel_test_count > 1:
            paralle_count = min(len(candidate_aliases), parallel_test_count)
            joblib.Parallel(n_jobs=paralle_count, require='sharedmem')\
                    (joblib.delayed(tool_parallel_test_exec)(ttoolalias) \
                        for ttoolalias in candidate_aliases)
        else:
            for tpos, ttoolalias in enumerate(candidate_aliases):
                found_a_failure = tool_parallel_test_exec(ttoolalias)
                if stop_on_failure and found_a_failure:
                    # @Checkpoint: Chekpointing for remaining tools
                    for rem_tool in list(testcases_by_tool.keys())[tpos+1:]:
                        checkpoint_handler.do_checkpoint(\
                                func_name=cp_func_name, \
                                taskid=cp_task_id, \
                                tool=rem_tool, \
                                opt_payload=meta_test_failedverdicts_outlog)
                    break
                                        
        if stop_on_failure:
            # Make sure the non executed test has the uncertain value (None)
            if len(meta_test_failedverdicts_outlog[0]) < len(meta_testcases):
                for meta_testcase in set(meta_testcases) - \
                                    set(meta_test_failedverdicts_outlog[0]):
                    meta_test_failedverdicts_outlog[0][meta_testcase] = \
                            common_mix.GlobalConstants.UNCERTAIN_TEST_VERDICT
                    meta_test_failedverdicts_outlog[1][meta_testcase] = \
                            common_matrices.OutputLogData.\
                                                    UNCERTAIN_TEST_OUTLOGDATA
                    
        ERROR_HANDLER.assert_true(len(meta_test_failedverdicts_outlog[0]) == \
                                                        len(meta_testcases), \
                            "Not all tests have a verdict reported", __file__)

        if fault_test_execution_matrix_file is not None:
            # Load or Create the matrix 
            fault_test_execution_matrix = common_matrices.ExecutionMatrix( \
                                filename=fault_test_execution_matrix_file, \
                                            non_key_col_list=meta_testcases)
            ERROR_HANDLER.assert_true(fault_test_execution_matrix.is_empty(), \
                                "matrix must be empty. Filename is:"
                                " "+fault_test_execution_matrix_file, __file__)
            failverdict2val = {
                common_mix.GlobalConstants.FAIL_TEST_VERDICT: \
                        fault_test_execution_matrix.getActiveCellDefaultVal(),
                common_mix.GlobalConstants.PASS_TEST_VERDICT: \
                            fault_test_execution_matrix.getInactiveCellVal(),
                common_mix.GlobalConstants.UNCERTAIN_TEST_VERDICT: \
                    fault_test_execution_matrix.getUncertainCellDefaultVal(),
            }
            cells_dict = {}
            for meta_testcase in meta_test_failedverdicts_outlog[0]:
                cells_dict[meta_testcase] = failverdict2val[\
                            meta_test_failedverdicts_outlog[0][meta_testcase]]

            fault_test_execution_matrix.add_row_by_key(self.FAULT_MATRIX_KEY, \
                                                cells_dict, serialize=True)

        if fault_test_execution_execoutput_file is not None:
            # Load or Create the data object 
            fault_test_execution_execoutput = common_matrices.OutputLogData( \
                                filename=fault_test_execution_execoutput_file)
            ERROR_HANDLER.assert_true(\
                            fault_test_execution_execoutput.is_empty(), \
                                        "outlog data must be empty", __file__)
            fault_test_execution_execoutput.add_data(\
                                    {self.PROGRAM_EXECOUTPUT_KEY: \
                                         meta_test_failedverdicts_outlog[1]}, \
                                                                serialize=True)


        # @Checkpoint: Finished
        detailed_exectime = {}
        for ttoolalias in testcases_by_tool.keys():
            tt = self.testcases_configured_tools[ttoolalias][self.TOOL_OBJ_KEY]
            detailed_exectime[ttoolalias] = (\
                        tt.get_checkpointer().get_execution_time(),\
                        tt.get_checkpointer().get_detailed_execution_time())

        checkpoint_handler.set_finished( \
                                    detailed_exectime_obj=detailed_exectime)

        if finish_destroy_checkpointer:
            checkpoint_handler.destroy()

        return meta_test_failedverdicts_outlog
    #~ def runtests()

    def get_candidate_tools_aliases(self, test_tool_type_list):
        candidate_tools_aliases = []
        for test_tool_type in test_tool_type_list:
            # validate the test_tool_types values
            ERROR_HANDLER.assert_true(\
                                TestToolType.is_valid(test_tool_type), \
                    "Invalid test tool type passed to test generation", \
                                                                __file__)
            for ttoolalias in self.testcases_configured_tools:
                if self.testcases_configured_tools[ttoolalias]\
                                    [self.TOOL_TYPE_KEY] == test_tool_type:
                    candidate_tools_aliases.append(ttoolalias) 
        return candidate_tools_aliases
    #~ def get_candidate_tools_aliases()

    def generate_tests (self, meta_criteria_tool_obj=None, \
                                exe_path_map=None, \
                                test_tool_type_list=None, \
                                max_time=None, \
                                test_generation_guidance_obj=None, \
                                parallel_testgen_count=1, \
                                restart_checkpointer=False, \
                                finish_destroy_checkpointer=True):
        """ This method should be used to generate the tests and must 
            always have a single instance running (it has single checkpoint
            file). 
            Note: The caller must explicitely destroy the checkpointer
            after this call succeed, to ensure that a sceduler will not
            re-execute this 
        :type meta_criteria_tool_obj:
        :param meta_criteria_tool_obj:

        :type exe_path_map:
        :param exe_path_map:

        :type test_tool_type_list:
        :param test_tool_type_list:
    
        :type \test_generation_guidance_obj:
        :param \test_generation_guidance_obj:
    
        :type \parallel_testgen_count:
        :param \parallel_testgen_count:

        :type restart_checkointer: bool
        :param restart_checkointer: Decide whether to discard checkpoint
                        and restart anew.

        :type finish_destroy_checkpointer: bool
        :param finish_destroy_checkpointer: Decide whether to automatically 
                        destroy the checkpointer when done or not
                        Useful is caller has a checkpointer to update. 
    
        :raises:
    
        :rtype:
        """
        # FIXME: Support test_generation_guidance_obj, then remove the code
        # bellow:
        ERROR_HANDLER.assert_true(test_generation_guidance_obj is None, \
                "FIXME: Must first implement support for test gen guidance")
        ERROR_HANDLER.assert_true(parallel_testgen_count <= 1, \
                "FIXME: Must first implement support for parallel test gen")
        #~ FXIMEnd

        # Check arguments Validity
        if exe_path_map is None:
            exe_path_map = self._get_default_exe_path_map()

        ERROR_HANDLER.assert_true(parallel_testgen_count > 0, \
                    "invalid parallel test generation count: {}. {}".format( \
                                    parallel_testgen_count, "must be >= 1"))
        if test_tool_type_list is None:
            candidate_tools_aliases = self.testcases_configured_tools.keys()
        else:
            ERROR_HANDLER.assert_true(len(test_tool_type_list) > 0,\
                                "Invalid test_tool_type_list passed (empty)", \
                                                                    __file__)
            candidate_tools_aliases = self.get_candidate_tools_aliases(\
                                                        test_tool_type_list)

        # @Checkpoint: create a checkpoint handler
        cp_func_name = "generate_tests"
        if test_tool_type_list is not None:
            for test_tool_type in sorted(test_tool_type_list, \
													key=lambda x: x.get_str()):
                cp_func_name += ":" + test_tool_type.get_str()
        cp_task_id = 1
        checkpoint_handler = CheckPointHandler(\
                                            self.get_checkpoint_state_object())
        if restart_checkpointer:
            checkpoint_handler.restart()
        if checkpoint_handler.is_finished():
            return

        # Generate
        for ttoolalias in candidate_tools_aliases:
            ttool = self.testcases_configured_tools[ttoolalias]\
                                                            [self.TOOL_OBJ_KEY]
            # Make sure to execute the right one
            if meta_criteria_tool_obj is None:
                if ttool.requires_criteria_instrumented():
                    continue
            else:
                if not ttool.requires_criteria_instrumented():
                    continue

            # Check whether already executed
            if checkpoint_handler.is_to_execute(func_name=cp_func_name, \
                                                taskid=cp_task_id, \
                                                tool=ttoolalias):

                # Actual Execution
                ttool.generate_tests(exe_path_map, \
                            meta_criteria_tool_obj=meta_criteria_tool_obj, \
                            max_time=max_time)

                # @Checkpoint: Checkpointing
                checkpoint_handler.do_checkpoint(func_name=cp_func_name, \
                                                taskid=cp_task_id, \
                                                tool=ttoolalias)

        # Invalidate any existing testcase info so it can be recomputed
        self._invalidate_testcase_info()

        # @Checkpoint: Finished
        detailed_exectime = {}
        for ttoolalias in candidate_tools_aliases:
            tt = self.testcases_configured_tools[ttoolalias][self.TOOL_OBJ_KEY]
            detailed_exectime[ttoolalias] = (\
                        tt.get_checkpointer().get_execution_time(),\
                        tt.get_checkpointer().get_detailed_execution_time())

        checkpoint_handler.set_finished( \
                                    detailed_exectime_obj=detailed_exectime)

        if finish_destroy_checkpointer:
            checkpoint_handler.destroy()
    #~ def generate_tests()
    
    def _compute_testcases_info(self, candidate_tool_aliases=None):
        meta_testcase_info_obj = TestcasesInfoObject()
        if candidate_tool_aliases is None:
            candidate_tool_aliases = self.testcases_configured_tools.keys()
        for ttoolalias in candidate_tool_aliases:
            ttool = \
                self.testcases_configured_tools[ttoolalias][self.TOOL_OBJ_KEY]
            tool_testcase_info = ttool.get_testcase_info_object()
            old2new_tests = {}
            for t_test in tool_testcase_info.get_tests_list():
                meta_t_key = DriversUtils.make_meta_element(t_test, ttoolalias)
                old2new_tests[t_test] = meta_t_key
            meta_testcase_info_obj.update_using(ttoolalias, old2new_tests, \
                                                            tool_testcase_info)
        return meta_testcase_info_obj
    #~ def _compute_testcases_info()
    
    def get_testcase_info_object(self, candidate_tool_aliases=None):
        tc_info = TestcasesInfoObject()
        tc_info.load_from_file(self.get_testcase_info_file(\
                                                       candidate_tool_aliases))
        return tc_info
    #~ def get_testcase_info_object()

    def get_testcase_info_file(self, candidate_tool_aliases=None):
        # Compute and write the testcase info if not present
        # only place where the meta info is written

        ## commented because many stages for test generation
        #if self._testcase_info_is_invalidated(): 
        self._compute_testcases_info(candidate_tool_aliases).write_to_file(\
                                    self._unchecked_get_testcase_info_file())
        return self._unchecked_get_testcase_info_file()
    #~ def get_testcase_info_file()

    def _unchecked_get_testcase_info_file(self):
        return self.testcases_info_file
    #~ def _unchecked_get_testcase_info_file():

    def _invalidate_testcase_info(self):
        if os.path.isfile(self._unchecked_get_testcase_info_file()):
            os.remove(self._unchecked_get_testcase_info_file())
    #~ def _invalidate_testcase_info()

    def _testcase_info_is_invalidated(self):
        return not os.path.isfile(self._unchecked_get_testcase_info_file())
    #~ def _testcase_info_is_invalidated()

    def _get_test_tool_out_folder(self, test_toolname, top_outdir=None):
        if top_outdir is None:
            top_outdir = self.tests_working_dir
        return os.path.join(top_outdir, test_toolname)
    #~ def _get_test_tool_out_folder()

    def _get_test_tool_checkpoint_files(self, test_toolname, \
                                                    top_checkpoint_dir=None):
        if top_checkpoint_dir is None:
            top_checkpoint_dir = self.checkpoints_dir
        return [os.path.join(top_checkpoint_dir, \
                            test_toolname+"_checkpoint.state"+suffix) \
                            for suffix in ("", ".backup")]
    #~ def _get_test_tool_checkpoint_files()
        
    def _get_checkpoint_files(self):
        top_checkpoint_dir = self.checkpoints_dir
        return [os.path.join(top_checkpoint_dir, \
                        "checkpoint.state"+suffix) \
                        for suffix in ("", ".backup")]
    #~ def _get_checkpoint_files()

    def get_test_tools_by_name(self, toolname):
        res = {}
        for alias, data in self.testcases_configured_tools.items():
            if data[self.TOOL_OBJ_KEY].get_toolname() == toolname:
                res[alias] = data[self.TOOL_OBJ_KEY]
        return res
    #~ def get_test_tools_by_name()

    def get_checkpoint_state_object(self):
        return self.checkpointer
    #~ def get_checkpoint_state_object()

    def has_checkpointer(self):
        return self.checkpointer is not None
    #~ def has_checkpointer()

    ######################################
    ############ OTHER FUNCTION ##########
    ######################################
    def get_flakiness_workdir(self):
        return self.flakiness_workdir
    #~ def get_flakiness_workdir()

    def check_get_flakiness(self, meta_testcases, repeat_count=2):
        """
            Check if tests have flakiness by running multiple times
            :return: The list of flaky tests
        """
        ERROR_HANDLER.assert_true(repeat_count > 1, "Cannot check flakiness"
                                    " with less than on repetition", __file__)

        if os.path.isdir(self.flakiness_workdir):
            shutil.rmtree(self.flakiness_workdir)
        os.mkdir(self.flakiness_workdir)

        outlog_files = [os.path.join(self.flakiness_workdir, \
                            str(of)+'-out.json') for of in range(repeat_count)]
        matrix_files = [os.path.join(self.flakiness_workdir, \
                            str(of)+'-mat.csv') for of in range(repeat_count)]
        
        def run(rep, test_list, hash_outlog):
            self.runtests(test_list, \
                        fault_test_execution_matrix_file=matrix_files[rep], \
                        fault_test_execution_execoutput_file=\
                                                            outlog_files[rep],\
                        with_output_summary=True, \
                        hash_outlog=hash_outlog)
        #~ def run()

        # Execute with repetition and get output summaries
        for rep in range(repeat_count):
            run(rep, meta_testcases, True)

        # get flaky tests list
        flaky_tests = set()
        fix_outdata = common_matrices.OutputLogData(filename=outlog_files[0])
        fix_outdata = list(fix_outdata.get_zip_objective_and_data())[0][1]
        for i in range(1, repeat_count):
            other_outdata = common_matrices.OutputLogData(\
                                                    filename=outlog_files[i])
            other_outdata = list(\
                            other_outdata.get_zip_objective_and_data())[0][1]
            for test, t_dat_fix in fix_outdata.items():
                if test in flaky_tests:
                    continue
                t_dat_other = other_outdata[test]
                if not common_matrices.OutputLogData.outlogdata_equiv(\
                                                    t_dat_fix, t_dat_other):
                    flaky_tests.add(test)

        ## cleanup
        for f in outlog_files + matrix_files:
            if os.path.isfile(f):
                os.remove(f)

        # get flaky tests outputs
        if len(flaky_tests) > 0:
            for rep in range(repeat_count):
                run(rep, list(flaky_tests), False)
            flaky_test_list_file = os.path.join(self.flakiness_workdir, \
                                                        "flaky_test_list.json")
            logging.warning("There were some flaky tests (see file {})"\
                                                .format(flaky_test_list_file))
            common_fs.dumpJSON(list(flaky_tests), flaky_test_list_file, \
                                                                pretty=True)
        return list(flaky_tests)
    #~ def check_get_flakiness()
#~ class MetaTestcaseTool




from __future__ import print_function
import os
import sys
import glob
import shutil
import logging
import resource
import random

import numpy as np

import muteria.common.fs as common_fs
import muteria.common.mix as common_mix
import muteria.common.matrices as common_matrices

import muteria.drivers.criteria as criteria
import muteria.controller.explorer as fd_structure

from muteria.repositoryandcode.codes_convert_support import CodeFormats
from muteria.drivers.testgeneration.base_testcasetool import BaseTestcaseTool
from muteria.drivers.testgeneration.testcases_info import TestcasesInfoObject
from muteria.drivers import DriversUtils
from muteria.drivers.testgeneration.testcase_formats.ktest.ktest \
                                                        import KTestTestFormat
from muteria.drivers.testgeneration.tools_by_languages.c.klee.klee \
                                                    import TestcasesToolKlee

ERROR_HANDLER = common_mix.ErrorHandler

class TestcasesToolSemu(TestcasesToolKlee):

    @classmethod
    def installed(cls, custom_binary_dir=None):
        """ Check that the tool is installed
            :return: bool reprenting whether the tool is installed or not 
                    (executable accessible on the path)
                    - True: the tool is installed and works
                    - False: the tool is not installed or do not work
        """
        for prog in ('klee-semu',):
            if custom_binary_dir is not None:
                prog = os.path.join(custom_binary_dir, prog)
            if not DriversUtils.check_tool(prog=prog, args_list=['--version'],\
                                                    expected_exit_codes=[0,1]):
                return False

        return KTestTestFormat.installed(custom_binary_dir=custom_binary_dir)
    #~ def installed()

    def __init__(self, *args, **kwargs):
        TestcasesToolKlee.__init__(self, *args, **kwargs)
        self.cand_muts_file = os.path.join(self.tests_working_dir, \
                                                        "cand_muts_file.txt")
        self.sm_mat_file = self.head_explorer.get_file_pathname(\
                            fd_structure.CRITERIA_MATRIX[criteria.TestCriteria\
                                                            .STRONG_MUTATION])
        self.mutants_by_funcs = None
    #~ def __init__()

    # SHADOW override
    def _get_default_params(self):
        bool_params = {
            '-ignore-solver-failures': None,
            '-allow-external-sym-calls': True, #None,
            '-posix-runtime': True, #None,
            '-dump-states-on-halt': True, #None,
            '-only-output-states-covering-new': None,
            '-use-cex-cache': True,
            # SEMu 
            '-semu-disable-statediff-in-testgen': None,
            '-semu-continue-mindist-out-heuristic': None,
            '-semu-use-basicblock-for-distance': None,
            '-semu-forkprocessfor-segv-externalcalls': True,
            '-semu-testsgen-only-for-critical-diffs': None,
            '-semu-consider-outenv-for-diffs': None,
        }
        key_val_params = {
            '-output-dir': self.tests_storage_dir,
            '-solver-backend': 'z3',
            '-max-solver-time': '300',
            '-search': 'bfs',
            '-max-memory': None,
            '-max-time': self.config.TEST_GENERATION_MAXTIME,
            '-libc': 'uclibc',
            '-max-sym-array-size': '4096',
            '-max-instruction-time': '10.',
            # SEMu 
            '-semu-mutant-max-fork': '0', #None,
            '-semu-checknum-before-testgen-for-discarded': '2', # None,
            '-semu-mutant-state-continue-proba': '0.25', #None,
            '-semu-precondition-length': '0', #None,
            '-semu-max-total-tests-gen': None,
            '-semu-max-tests-gen-per-mutant': '5', # None,
            '-semu-loop-break-delay': '120.0',
        }
        key_val_params.update({
                                })
        # XXX: set muts cand list
        if os.path.isfile(self.sm_mat_file):
            key_val_params['-semu-candidate-mutants-list-file'] = \
                                                        self.cand_muts_file
        return bool_params, key_val_params
    #~ def _get_default_params()
    
    def _call_generation_run(self, runtool, args):
        # use mutants_by_funcs to reorganize target mutants for scalability

        max_mutant_count_per_cluster = self.get_value_in_arglist(args, \
                                        'DRIVER_max_mutant_count_per_cluster')
        if max_mutant_count_per_cluster is None:
            # TODO: compute max_mutant_count_per_cluster based on max memory
            max_mutant_count_per_cluster = 100
        else:
            max_mutant_count_per_cluster = float(max_mutant_count_per_cluster)
            self.remove_arg_and_value_from_arglist(args, \
                                        'DRIVER_max_mutant_count_per_cluster')

        cand_mut_file_bak = self.cand_muts_file + '.bak'
        mut_list = []
        with open(self.cand_muts_file) as f:
            for m in f:
                m = m.strip()
                ERROR_HANDLER.assert_true(m.isdigit(), "Invalid mutant ID", \
                                                                    __file__)
                mut_list.append(m)
        if self.mutants_by_funcs is None:
            random.shuffle(mut_list)
        else:
            # make the mutants to be localized in same function as much as
            # possible
            # XXX Here the mutant ID are NOT meta but simple IDs
            this_mutants_by_funcs = {}
            mut_to_func = {}
            for f, m_set in self.mutants_by_funcs.items():
                this_mutants_by_funcs[f] = m_set & set(mut_list)
                for m in this_mutants_by_funcs[f]:
                    mut_to_func[m] = f
            func_to_rank_dec = list(this_mutants_by_funcs.keys())
            func_to_rank_dec.sort(key=lambda x: len(this_mutants_by_funcs[x]),\
                                                                reverse=True)
            func_to_rank_dec = {f: r for r, f in enumerate(func_to_rank_dec)}
            mut_list.sort(key=lambda x: func_to_rank_dec[mut_to_func[x]])
        nclust = int(len(mut_list) / max_mutant_count_per_cluster)
        if len(mut_list) != max_mutant_count_per_cluster * nclust:
            nclust += 1
        clusters = np.array_split(mut_list, nclust)

        # update max-time
        if len(clusters) > 1:
            cur_max_time = float(self.get_value_in_arglist(args, 'max-time'))
            self.set_value_in_arglist(args, 'max-time', \
                                    str(max(1, cur_max_time / len(clusters))))
        
        shutil.move(self.cand_muts_file, cand_mut_file_bak)

        c_dirs = []
        for c_id, clust in enumerate(clusters):
            logging.debug("SEMU: targeting mutant cluster {}/{} ...".format(\
                                                        c_id+1, len(clusters)))
            with open(self.cand_muts_file, 'w') as f:
                for m in clust:
                    f.write(m+'\n')

            super(TestcasesToolSemu, self)._call_generation_run(runtool, args)
            
            c_dir = os.path.join(os.path.dirname(self.tests_storage_dir), \
                                                                    str(c_id))
            shutil.move(self.tests_storage_dir, c_dir)
            c_dirs.append(c_dir)

        os.mkdir(self.tests_storage_dir)
        for c_dir in c_dirs:
            shutil.move(c_dir, self.tests_storage_dir)

        shutil.move(cand_mut_file_bak, self.cand_muts_file)
    #~ def _call_generation_run()

    def _get_tool_name(self):
        return 'klee-semu'
    #~ def _get_tool_name()
    
    def _get_input_bitcode_file(self, code_builds_factory, rel_path_map, \
                                                meta_criteria_tool_obj=None):
        # XXX: get the meta criterion file from MART.
        mutant_gen_tool_name = 'mart'
        mut_tool_alias_to_obj = \
                            meta_criteria_tool_obj.get_criteria_tools_by_name(\
                                                        mutant_gen_tool_name)

        if len(mut_tool_alias_to_obj) == 0:
            logging.warning(\
                        'SEMu requires Mart to generate mutants but none used')

        ERROR_HANDLER.assert_true(len(mut_tool_alias_to_obj) == 1, \
                                "SEMu supports tests generation from"
                                "a single .bc file for now (todo).", __file__)

        t_alias2metamu_bc = {}
        t_alias2mutantInfos = {}
        for alias, obj in mut_tool_alias_to_obj.items():
            dest_bc = rel_path_map[list(rel_path_map)[0]]+'.bc'
            shutil.copy2(obj.get_test_gen_metamutant_bc(), dest_bc)
            t_alias2metamu_bc[alias] = dest_bc
            t_alias2mutantInfos[alias] = obj.get_criterion_info_object(None)

        # XXX: get mutants ids by functions
        self.mutants_by_funcs = {}
        single_alias = list(t_alias2mutantInfos)[0]
        single_tool_obj = t_alias2mutantInfos[single_alias]
        for mut in single_tool_obj.get_elements_list():
            func = single_tool_obj.get_element_data(mut)[\
                                                        'mutant_function_name']
            #meta_mut = DriversUtils.make_meta_element(mut, single_alias)
            if func not in self.mutants_by_funcs:
                self.mutants_by_funcs[func] = set()
            self.mutants_by_funcs[func].add(mut) #meta_mut)

        # XXX: get candidate mutants list
        if os.path.isfile(self.sm_mat_file):
            sm_mat = common_matrices.ExecutionMatrix(filename=self.sm_mat_file)
            mut2killing_tests = sm_mat.query_active_columns_of_rows()
            alive_muts = [m for m, k_t in mut2killing_tests.items() \
                                                            if len(k_t) == 0]
            with open(self.cand_muts_file, 'w') as f:
                for meta_m in alive_muts:
                    t_alias, m = DriversUtils.reverse_meta_element(meta_m)
                    if t_alias in t_alias2metamu_bc: # There is a single one
                        f.write(str(m)+'\n')

        return t_alias2metamu_bc[list(t_alias2metamu_bc)[0]]
    #~ def _get_input_bitcode_file()

    def requires_criteria_instrumented(self):
        return True
    #~ def requires_criteria_instrumented()
#~ class TestcasesToolSemu


#
# [LICENCE]
#
""" Default criteria test execution optimizer class
"""

from __future__ import print_function
import os
import sys
import copy
import logging

import muteria.common.mix as common_mix
import muteria.common.matrices as common_matrices

import muteria.controller.explorer as explorer
from muteria.drivers import DriversUtils
from muteria.drivers.criteria import TestCriteria

from muteria.drivers.optimizers.criteriatestexecution.\
                                base_criteria_test_execution_optimizer \
                                    import BaseCriteriaTestExecutionOptimizer

from muteria.drivers.optimizers.testexecution.tools.default \
                                                import TestExecutionOptimizer

ERROR_HANDLER = common_mix.ErrorHandler

class CriteriaTestExecutionOptimizer(BaseCriteriaTestExecutionOptimizer):

    #######################################################################
    ##################### Methods implemented ############################
    #######################################################################

    @classmethod
    def installed(cls, custom_binary_dir=None):
        """ Check that the tool is installed
            :return: bool reprenting whether the tool is installed or not 
                    (executable accessible on the path)
                    - True: the tool is installed and works
                    - False: the tool is not installed or do not work
        """
    #~ def installed()

    def reset (self, toolalias, test_objective_list, test_list, **kwargs):
        """ Reset the optimizer
        """
        self.test_objective_ordered_list = copy.deepcopy(test_objective_list)
        self.pointer = 0
        self.test_objective_to_test_execution_optimizer = {
            to: TestExecutionOptimizer(self.config, self.explorer) \
            for to in self.test_objective_ordered_list
        }

        # get the test list per test objectives, based on the matrix
        opt_by_criterion = self._get_optimizing_criterion()
        matrix_file = self.explorer.get_existing_file_pathname(\
                                explorer.TMP_CRITERIA_MATRIX[opt_by_criterion])
        test_list_per_test_objective = \
                            self._get_test_objective_tests_from_matrix(\
                                                    matrix_file=matrix_file)

        # Update test list
        for to, teo in self.test_objective_to_test_execution_optimizer.items():
            alias_to = DriversUtils.make_meta_element(to, toolalias)
            ERROR_HANDLER.assert_true(\
                                    alias_to in test_list_per_test_objective, \
                                    "Bug: test objective missing("+str(to)+')')
            teo.reset(None, test_list_per_test_objective[alias_to], \
                                                            disable_reset=True)
    #~ def reset()

    ##### Private methods #####
    
    def _get_optimizing_criterion(self):
        return TestCriteria.WEAK_MUTATION
    #~ def _get_optimizing_criterion()

    def _get_test_objective_tests_from_matrix (self, matrix_file):
        matrix = common_matrices.ExecutionMatrix(filename=matrix_file)
        res = matrix.query_active_columns_of_rows()
        return res
    #~ def _get_test_objective_tests_from_matrix ()
#~ class CriteriaTestExecutionOptimizer




#
# [LICENCE]
#
""" Default test execution optimizer class
"""

from __future__ import print_function
import os
import sys
import copy
import logging

import muteria.common.mix as common_mix

from muteria.drivers.optimizers.testexecution.base_test_execution_optimizer \
                                            import BaseTestExecutionOptimizer

ERROR_HANDLER = common_mix.ErrorHandler

class TestExecutionOptimizer(BaseTestExecutionOptimizer):

    #######################################################################
    ##################### Methods implemented ############################
    #######################################################################

    @classmethod
    def installed(cls, custom_binary_dir=None):
        """ Check that the tool is installed
            :return: bool reprenting whether the tool is installed or not 
                    (executable accessible on the path)
                    - True: the tool is installed and works
                    - False: the tool is not installed or do not work
        """
    #~ def installed()

    def reset (self, toolalias, test_list, disable_reset=False, **kwargs):
        """ Reset the optimizer
        """
        ERROR_HANDLER.assert_true(not self.reset_disabled, "reset is disabled")
        self.test_ordered_list = copy.deepcopy(test_list)
        self.pointer = 0

        if disable_reset and not self.reset_disabled:
            self.reset_disabled = True
    #~ def reset()
#~ class TestExecutionOptimizer



from __future__ import print_function

import os
import sys
import re
import shutil
import imp
from distutils.spawn import find_executable

import muteria.common.mix as common_mix

from muteria.drivers import DriversUtils

ERROR_HANDLER = common_mix.ErrorHandler

class KTestTestFormat(object):
    
    @classmethod
    def installed(cls, custom_binary_dir=None):
        """ Check that the tool is installed
            :return: bool reprenting whether the tool is installed or not 
                    (executable accessible on the path)
                    - True: the tool is installed and works
                    - False: the tool is not installed or do not work
        """
        for prog in (cls.tool,):
            if custom_binary_dir is not None:
                prog = os.path.join(custom_binary_dir, prog)
            if not DriversUtils.check_tool(prog=prog, args_list=['--version'],\
                                                    expected_exit_codes=[1]):
                return False
        return True
    #~ def installed()

    @classmethod
    def get_test_replay_tool(cls, custom_replay_tool_binary_dir=None):
        if custom_replay_tool_binary_dir is None:
            kr_file = find_executable(cls.tool)
            ERROR_HANDLER.assert_true(kr_file is not None, \
                        "Could not fine test replay tool on path", __file__)
        else:
            kr_file = os.path.join(custom_replay_tool_binary_dir, cls.tool)
            ERROR_HANDLER.assert_true(os.path.isfile(kr_file), \
                                "replay tool not found in custom_binary_dir", \
                                                                    __file__)
        return kr_file
    #~ def get_test_replay_tool()

    @classmethod
    def _get_replay_prog_args(cls, executable_file, test_file, \
                                        custom_replay_tool_binary_dir=None):
        prog = cls.tool
        if custom_replay_tool_binary_dir is not None:
            prog = os.path.join(custom_replay_tool_binary_dir, prog)
            ERROR_HANDLER.assert_true(os.path.isfile(prog), \
                            "The tool {} is missing from the specified dir {}"\
                            .format(cls.tool, custom_replay_tool_binary_dir), \
                                                                    __file__)

        args = [executable_file, test_file]
        return prog, args
    #~ def _get_replay_prog_args()

    @classmethod
    def execute_test(cls, executable_file, test_file, env_vars, timeout=None, \
                    collected_output=None, custom_replay_tool_binary_dir=None):

        prog, args = cls._get_replay_prog_args(executable_file, test_file, \
                                                custom_replay_tool_binary_dir)

        # klee-replay may create files or dir. in KLEE version with LLVM-3.4,
        # those are created in a temporary dir set as <cwd>.temps
        # XXX XXX. make sure each test has its own
        test_work_dir = test_file+".execdir"
        klee_replay_temps = test_work_dir + '.temps'
        if not os.path.isdir(test_work_dir):
            os.mkdir(test_work_dir)
        if os.path.isdir(klee_replay_temps):
            try:
                shutil.rmtree(klee_replay_temps)
            except PermissionError:
                cls._dir_chmod777(klee_replay_temps)
                shutil.rmtree(klee_replay_temps)

        # XXX Execution setup
        tmp_env = os.environ.copy()
        if env_vars is not None:
            #for e, v in env_vars.items():
            #    tmp_env[e] = v
            tmp_env.update(env_vars)

        timeout_return_codes = cls.timedout_retcodes + \
                                        DriversUtils.EXEC_TIMED_OUT_RET_CODE

        if timeout is not None:
            tmp_env['KLEE_REPLAY_TIMEOUT'] = str(timeout)
            kt_over = 5 # 1second
            timeout += kt_over
        
        # XXX Execute the ktest
        if collected_output is not None:
            retcode, out, err = DriversUtils.execute_and_get_retcode_out_err(\
                                prog=prog, args_list=args, env=tmp_env, \
                                timeout=timeout, timeout_grace_period=5, \
                                merge_err_to_out=True, cwd=test_work_dir)
            out = cls._remove_output_noise(out)
            collected_output.extend((retcode, out, \
                                            (retcode in timeout_return_codes)))
        else:
            retcode, out, err = DriversUtils.execute_and_get_retcode_out_err(\
                                prog=prog, args_list=args, env=tmp_env, \
                                timeout=timeout, timeout_grace_period=5, \
                                                out_on=False, err_on=False, \
                                                cwd=test_work_dir)

        # XXX: Go back to previous CWD
        if os.path.isdir(klee_replay_temps):
            try:
                shutil.rmtree(klee_replay_temps)
            except PermissionError:
                cls._dir_chmod777(klee_replay_temps)
                shutil.rmtree(klee_replay_temps)

        if retcode in timeout_return_codes + \
                                    DriversUtils.EXEC_SEGFAULT_OUT_RET_CODE:
            verdict = common_mix.GlobalConstants.FAIL_TEST_VERDICT
        else:
            verdict = common_mix.GlobalConstants.PASS_TEST_VERDICT

        return verdict
    #~ def execute_test()

    @staticmethod
    def _dir_chmod777(dirpath):
        try:
            for root_, dirs_, files_ in os.walk(dirpath):
                for sub_d in dirs_:
                    os.chmod(os.path.join(root_, sub_d), 0o777)
                for f_ in files_:
                    os.chmod(os.path.join(root_, f_), 0o777)
        except PermissionError:
            ret,out,_ = DriversUtils.execute_and_get_retcode_out_err('sudo', \
                                            ['chmod', '777', '-R', dirpath])
            ERROR_HANDLER.assert_true(ret == 0, \
                        "'sudo chmod 777 -R "+dirpath+"' failed (returned "+\
                                        str(ret)+"), error: "+out, __file__)
    #~ def _dir_chmod777()
    
    @classmethod
    def get_replay_test_wrapper_str(cls, exe_env_var, ktest_env_var, \
                                                            timeout_env_var, \
                                        custom_replay_tool_binary_dir=None):
        # XXX: This is used by shadow to replay generated tests through
        # executing base dev test used to generate them.
        # TODO: Let shadow store the mapping between dev test and gen test
        # TODO: Use the mapping to run tests using the returned wrapper by this function
        prog, args = cls._get_replay_prog_args('${}'.format(exe_env_var), \
                                                '${}'.format(ktest_env_var), \
                                               custom_replay_tool_binary_dir)
        
        python_code = ';'.join(['import sys', \
                'from muteria.drivers.testgeneration' \
                    + '.testcase_formats.ktest.ktest import KTestTestFormat', \
                'r = KTestTestFormat._remove_output_noise(sys.stdin.read())', \
                    'print(r)'])

        bash_timeout_retcode = os.system('timeout 0.1 sleep 1')

        ret_str = "#! /bin/bash\n\n"
        ret_str += "set -u\nset -o pipefail\n\n"
        ret_str += "export KLEE_REPLAY_TIMEOUT={}\n".format(timeout_env_var)
        ret_str += " ".join([prog] + args) + ' 2>&1 | {} -c "{}"\n'.format(\
                                                sys.executable, python_code)
        ret_str += "exit_code=$?\n"
        ret_str += '{} -c "exit(not($exit_code in {}))" && exit_code={}\n'\
                                                .format(sys.executable, \
                                                    cls.timedout_retcodes, \
                                                    bash_timeout_retcode)
        ret_str += "exit $exit_code\n"
        return ret_str
    #~ def get_replay_test_wrapper_str()

    timedout_retcodes = (88,) # taken from klee_replay source code
        
    tool = 'klee-replay'

    # Newer version (after klee github commit 88bb205)
    # (88bb205e422ee2aaf75594e4e314b21f77f219e3)
    #clean_everything_regex = re.compile("^KLEE-REPLAY: ")
    #clean_part_regex = re.compile("(" + "|".join(["^KLEE-REPLAY: NOTE: ",\
    #                                    "^KLEE-REPLAY: WARNING: ",\
    #                                    "^KLEE-REPLAY: klee_warning: ",\
    #                                    "^KLEE-REPLAY: klee_warning_once: ",\
    #                                    "^KLEE-REPLAY: klee_assume",\
    #                                    ]) + ")")

    # Older version (before klee github commit 88bb205)
    clean_everything_regex = re.compile("(" + "|".join([\
                        "^EXIT STATUS: .* \\([0-9]+\\s+seconds\\)$", \
                        ""+tool+": EXIT STATUS: .* \\([0-9]+\\s+seconds\\)$", \
                        ""+tool+": received signal [0-9]+\\s+. "+\
                                        "Killing monitored process\\(es\\)$", \
                        "^note: (pty|pipe) (master|slave): ",\
                        ""+tool+": PTY (MASTER|SLAVE): EXIT STATUS: ", \
                        "^warning: check_file .*: .* "+\
                                        "mismatch: [0-9]+ [vV][sS] [0-9]+$", \
                        "^RUNNING GDB: /usr/bin/gdb --pid [0-9]+ -q --batch", \
                        "^TIMEOUT: ATTEMPTING GDB EXIT$", \
                        #"^ERROR: ", \
                        #"^Error: (executable|chroot:) ", \
                        #"^klee_range(\(|:)", \
                        #"^make_symbolic mismatch, different sizes: ", \
                        #"^WARNING: ", \
                        #"^rootdir: ", \
                        #""+tool+": error: input file ", \
                        ""+tool+": TEST CASE: ", \
                        ""+tool+": ARGS: ", \
                        ]) + ")")
    clean_part_regex = re.compile(("(" + "|".join([\
                        "^EXIT STATUS: .* \\([0-9]+\\s+seconds\\)$", \
                        ""+tool+": EXIT STATUS: .* \\([0-9]+\\s+seconds\\)$", \
                        ""+tool+": received signal [0-9]+\\s+. "+\
                                        "Killing monitored process\\(es\\)$", \
                        "^note: (pty|pipe) (master|slave): ",\
                        ""+tool+": PTY (MASTER|SLAVE): EXIT STATUS: ", \
                        "^warning: check_file .*: .* "+\
                                        "mismatch: [0-9]+ [vV][sS] [0-9]+$", \
                        "^RUNNING GDB: /usr/bin/gdb --pid [0-9]+ -q --batch", \
                        "^TIMEOUT: ATTEMPTING GDB EXIT$", \
                        ]) + ")"))

    @classmethod
    def _remove_output_noise(cls, out, clean_everything=True):
        res = []
        if clean_everything:
            regex = cls.clean_everything_regex
        else:
            regex = cls.clean_part_regex

        if len(out) > 0 and out[-1] == '\n':
            out = out[:-1]
            last_char = "\n" 
        else:
            last_char = ""

        for line in out.encode('utf-8', 'backslashreplace').splitlines():
            line = line.decode('utf-8', 'backslashreplace')
            if regex.search(line) is None:
                # None is matched
                res.append(line)

        res = '\n'.join(res) + last_char

        return res
    #~ def _remove_output_noise()

    ktest_extension = '.ktest'

    @classmethod
    def ktest_fdupes(cls, *args, custom_replay_tool_binary_dir=None):
        """
        This function computes the fdupes of the klee ktest directories 
        and ktest files given as arguments. 
        It requires that the files and directories passed as arguments exist

        :param *args: each argument is either a file or a directory that exists

        :return: returns two values: 
                - The first is a python list of tuples. 
                    each tuple represents files that are duplicate with each 
                    other and ranked by their age (modified time) the oldest 
                    first (earliest modified to latest modified).
                - The second is the list of files that are not valid
                    ktest files.
        """
        # import ktest
        ktt_dir = os.path.dirname(cls.get_test_replay_tool(
                                        custom_replay_tool_binary_dir=\
                                                custom_replay_tool_binary_dir))
        ktest_tool = imp.load_source("ktest-tool", \
                                        os.path.join(ktt_dir, 'ktest-tool'))

        ret_fdupes = []
        invalid = []
        file_set = set()
        for file_dir in args:
            if os.path.isfile(file_dir):
                file_set.add(file_dir)
            elif os.path.isdir(file_dir):
                # get ktest files recursively
                for root, directories, filenames in os.walk(file_dir):
                    for filename in filenames:
                        if filename.endswith(cls.ktest_extension):
                            file_set.add(os.path.join(root, filename))
            else:
                ERROR_HANDLER.error_exit(\
                        "Invalid file or dir passed (inexistant): "+file_dir, \
                                                                    __file__)

        # apply fdupes: load all ktests and strip the non uniform data 
        # (.bc file used) then compare the remaining data
        kt2used_dat = {}
        for kf in file_set:
            try:
                b = ktest_tool.KTest.fromfile(kf)
                kt2used_dat[kf] = (b.args[1:], b.objects)
            except:
                invalid.append(kf)

        # do fdupes
        dup_dict = {}
        keys = list(kt2used_dat.keys())
        for ktest_file in keys:
            if ktest_file in kt2used_dat:
                ktest_file_dat = kt2used_dat[ktest_file]
                del kt2used_dat[ktest_file]
                for other_file in kt2used_dat:
                    if kt2used_dat[other_file] == ktest_file_dat:
                        if ktest_file not in dup_dict:
                            dup_dict[ktest_file] = []
                        dup_dict[ktest_file].append(other_file)
                if ktest_file in dup_dict:
                    for dup_of_kt_file in dup_dict[ktest_file]:
                        del kt2used_dat[dup_of_kt_file]

        # Finilize
        for ktest_file in dup_dict:
            tmp = [ktest_file] + dup_dict[ktest_file]
            # sort by decreasing modified age
            tmp.sort(key=lambda x: os.path.getmtime(x))
            ret_fdupes.append(tuple(tmp))

        return ret_fdupes, invalid
    #~ def ktest_fdupes()
#~ class KTestTestFormat




from __future__ import print_function

import sys
import os
import argparse

from _version import __version__, _framework_name

import muteria.common.mix as common_mix

import muteria.configmanager.configurations as configurations
import muteria.configmanager.helper as configs_helper
from muteria.controller.main_controller import MainController

ERROR_HANDLER = common_mix.ErrorHandler


class CliUserInterface(object):
    @classmethod
    def _cmd_load(cls):
        """ Return a dict of conf file path, lang,... and other conf
            The required depend on the mode used (TODO: define required by modes)
        """
        parser = argparse.ArgumentParser()
        parser.add_argument('-v', '--version', action='version',
                                    version='\n{} {}\n'.format(_framework_name, 
                                                                  __version__))
        parser.add_argument("--config", help="Config file")
        parser.add_argument("--lang", help="Programming Language")
        subparsers = parser.add_subparsers(dest='command')
        parser_restore = subparsers.add_parser('restore', \
                                    help="Restore the repository to the"
                                        " initial state of the source files")
        parser_restore.add_argument("--asinitial", \
                                                help="remove also added files")

        parser_view = subparsers.add_parser('view', \
                                    help="View the output directory"
                                        " files and folders")
        parser_view.add_argument('--results', action='store_true', \
                                    help='Print the result folder location')

        parser_internal = subparsers.add_parser('internal', \
                                    help="Get informations of the"
                                        " framework such as tool list, ...")
        parser_internal.add_argument("--languages", action='store_true', \
                                help="Get the languages supported partially")

        parser_run = subparsers.add_parser('run', \
                                    help="Execute the tass specified in the"
                                        " configuration file")
        parser_run.add_argument("--cleanstart", action='store_true', \
                                            help="Clear out dir and restart")

        if len(sys.argv)==1:
            parser.print_help(sys.stderr)
            sys.exit(1)

        args = parser.parse_args()
        return args
    #~ def _cmd_load()

    @classmethod
    def get_cmd_raw_conf(cls):
        """ Get CLI raw configuration by parsing command line arguments
        """
        args = cls._cmd_load()

        ERROR_HANDLER.assert_true(args.config, "Must specify config", __file__)
        ERROR_HANDLER.assert_true(os.path.isfile(args.config), \
                                            "Config file inexistant", __file__)
        conf_file = args.config
        cfg_obj = configs_helper.ConfigurationHelper()

        ERROR_HANDLER.assert_true(args.lang, "Must specify language", __file__)
        lang = args.lang

        raw_conf = cfg_obj.get_extend_file_raw_conf(conf_file, lang)

        if args.command == 'run':
            if args.cleanstart:
                raw_conf['EXECUTION_CLEANSTART'] = True
            raw_conf['RUN_MODE'] = configurations.SessionMode.EXECUTE_MODE
        elif args.command == 'restore':
            raw_conf['RUN_MODE'] = \
                                configurations.SessionMode.RESTORE_REPOS_MODE
        elif args.command == 'view':
            raw_conf['RUN_MODE'] = configurations.SessionMode.VIEW_MODE
        elif args.command == 'internal':
            raw_conf['RUN_MODE'] = configurations.SessionMode.INTERNAL_MODE
        elif args.command == 'customexec':
            raw_conf['RUN_MODE'] = \
                            configurations.SessionMode.CUSTOM_EXECUTION_MODE
        else:
            ERROR_HANDLER.error_exit("must specify a command."
                            " use --help to see available commands", __file__)
        return raw_conf
    #~ def get_cmd_raw_conf()

    @classmethod
    def cmd_main(cls):
        # XXX Parse the arguments and get the raw configs
        cmd_raw_conf = cls.get_cmd_raw_conf()
        
        # Call raw_config_main with the created raw config
        ctrl = MainController()
        ctrl.raw_config_main(raw_config=cmd_raw_conf)
    #~ def cmd_main()

#~ class CliUserInterface

def main():
    CliUserInterface.cmd_main()

if __name__ == "__main__":
    main()
